"""
Help command and utils.
"""

from pax25.applications.parsers import string_parser
from pax25.applications.router import CommandContext, CommandRouter, CommandSpec
from pax25.applications.utils import build_columns, send_message
from pax25.services.connection.connection import Connection


def build_help_entry(spec: CommandSpec[object]) -> str:
    """
    Build a help entry from a spec.
    """
    entry = f"Topic: {spec.command}\r"
    if spec.aliases:
        entry += f"Aliases: {','.join(spec.aliases)}\r"
    return entry + "\r" + spec.help


class Help:
    """
    Help system. Initiate it with your command router to automatically build a help
    index for all commands.
    """

    def __init__(self, command_router: CommandRouter) -> None:
        """
        Initialize the help system.
        """
        self.command_router = command_router

    def send_index(self, connection: Connection) -> None:
        """
        Send an index of all commands and topics.
        """
        command_list: list[str] = []
        header = "Type 'help topic' where topic is one of:\r"
        for command in sorted(self.command_router.command_set):
            [spec] = self.command_router.command_store[command]
            command_list.append(spec.command)
        send_message(connection, header + "\r".join(build_columns(command_list)))

    def run_help(self, connection: Connection, context: CommandContext[str]) -> None:
        """
        Runs the 'help' command. Executed by the command router if we added the spec
        to it.
        """
        lookup = context.args.upper()
        if not lookup:
            self.send_index(connection)
            return
        try:
            candidates = self.command_router.command_store[lookup]
        except KeyError:
            send_message(
                connection,
                f"No help entry found for {repr(context.args)}. Type "
                "'help' for a list of entries.",
            )
            return
        if len(candidates) > 1:
            possibilities = sorted(entry.command for entry in candidates)
            send_message(
                connection,
                "Ambiguous topic. "
                f"Did you mean one of these?: {', '.join(possibilities)}",
            )
            return
        [spec] = candidates
        send_message(connection, build_help_entry(spec))

    @property
    def spec(self) -> CommandSpec[str]:
        """
        A CommandSpec that can be
        :return:
        """
        return CommandSpec(
            command="help",
            aliases=("?",),
            function=self.run_help,
            help="Provides information on various commands. Type 'help' on "
            "its own to get a listing of commands, or 'help command' where command "
            "is a command to get help on.",
            parser=string_parser,
        )



"""
Connection service. Keeps track of the connection table, instantiates connections,
performs other housekeeping around them.
"""

import logging
from collections import defaultdict
from typing import TYPE_CHECKING, Any

from pax25.applications import BaseApplication
from pax25.applications.types import S
from pax25.ax25.address import Address, AddressHeader, Route
from pax25.ax25.constants import UFrameType
from pax25.ax25.control import Unnumbered
from pax25.ax25.frame import Frame
from pax25.ax25.matchers import (
    U_FRAME_CHECKS,
    MatchCall,
    check_all,
    repeats_completed,
)
from pax25.ax25.utils import reply_digipeaters
from pax25.exceptions import ConfigurationError
from pax25.interfaces import Interface
from pax25.services.connection.connection import (
    Connection,
    ConnectionKey,
    connection_key,
)
from pax25.types import ConnectionSettings
from pax25.utils import smart_clone

if TYPE_CHECKING:  # pragma: no cover
    from pax25.station import Station


logger = logging.getLogger(__name__)


type ApplicationMap = defaultdict[str, dict[Address, BaseApplication[Any]]]


def default_connection_settings() -> "ConnectionSettings":
    """
    Default parameters for parameter settings.
    """
    return {
        "retries": 10,
        "retry_interval": 10000,
        "reception_status_delay": 2000,
        "connection_check_interval": 30_000,
    }


class ConnectionService:
    """
    Connection handling service.
    """

    def __init__(
        self,
        station: "Station",
        settings: ConnectionSettings | None,
    ) -> None:
        """
        Initialize the connection service.
        """
        self.station = station
        self._settings = default_connection_settings()
        if settings:
            self._settings.update(settings)
        # We keep track of the addresses we are internally watching for here, such
        # as contacts to our internally registered applications.
        self._internal_addresses: dict[Address, set[Interface[Any]]] = {}
        self.table: dict[ConnectionKey, Connection] = {}
        self._application_map: ApplicationMap = defaultdict(dict)
        self.station.frame_router.register_matcher(
            "connection",
            MatchCall(matcher=self.connection_matcher, notify=self.inbound_connect),
        )
        self.station.frame_router.register_matcher(
            "disconnect_mode",
            MatchCall(
                matcher=self.disconnect_mode_matcher,
                notify=self.disconnect_mode_sender,
            ),
        )

    @property
    def settings(self) -> ConnectionSettings:
        """
        Get the default tunable parameters for connected sessions.
        """
        return smart_clone(self._settings)

    async def reload_settings(self, settings: ConnectionSettings) -> None:
        """
        Reload the settings for connections.
        """
        old_settings = self._settings
        # Makes a clone.
        result = self.settings
        result.update(settings)
        if result == self._settings:
            return
        self._settings = settings
        # OK, looks like we need to reload all the connections, too.
        for value in self.table.values():
            # We don't want to change any connections that had anything special applied.
            if value.settings == old_settings:
                await value.reload_settings(settings)

    @property
    def application_map(self) -> ApplicationMap:
        """
        Return a copy of the application map.
        """
        return smart_clone(self._application_map)

    def connection_matcher(self, frame: Frame, interface: Interface) -> bool:
        """
        Matches incoming connection frames.
        """
        if interface.type == "Loopback":
            # Ignore all loopback connection frames.
            return False
        if not check_all(
            repeats_completed, U_FRAME_CHECKS[UFrameType.SET_ASYNC_BALANCED_MODE]
        )(frame, interface):
            return False
        return frame.route.dest.address in self._application_map.get(interface.name, [])

    def disconnect_mode_matcher(self, frame: Frame, interface: Interface) -> bool:
        """
        Checks if we have a frame that's addressed to us like we have a connection open,
        but no such connection exists.
        """
        if interface.type == "Loopback":
            return False
        if not repeats_completed(frame, interface):
            return False
        if frame.route.dest.address not in self._application_map.get(
            interface.name,
            [],
        ):
            return False
        if U_FRAME_CHECKS[UFrameType.SET_ASYNC_BALANCED_MODE](frame, interface):
            return False
        if U_FRAME_CHECKS[UFrameType.UNNUMBERED_INFORMATION](frame, interface):
            return False
        if U_FRAME_CHECKS[UFrameType.DISCONNECT_MODE](frame, interface):
            logger.warning(
                "Received Disconnect Mode frame when we weren't in a connection. "
                "There is likely a duplicate address on this network! (%s)",
                frame.route.dest.address,
            )
            return False
        first_party = frame.route.src.address
        second_party = frame.route.dest.address
        return connection_key(first_party, second_party, interface) not in self.table

    def disconnect_mode_sender(self, frame: Frame, interface: Interface) -> None:
        """
        Send a disconnect mode frame in response to a frame.
        """
        response_frame = Frame(
            route=Route(
                src=AddressHeader(address=frame.route.dest.address),
                dest=AddressHeader(address=frame.route.src.address),
                digipeaters=reply_digipeaters(frame.route.digipeaters),
            ),
            control=Unnumbered(
                frame_type=UFrameType.DISCONNECT_MODE,
                poll_or_final=True,
            ),
            pid=None,
        )
        self.station.frame_router.send_frame(interface, response_frame)

    def add_app(
        self,
        app: type[BaseApplication[S]],
        *,
        interfaces: list[str],
        application_name: str | None = None,
        station_name: str | None = None,
        ssid: int | None = None,
        settings: S,
    ) -> BaseApplication[S]:
        """
        Registers a new application to the specified interfaces.
        """
        station_name = (station_name if station_name else self.station.name).upper()
        application_name = (
            application_name if application_name is not None else app.__name__
        )
        application_instance = app(
            name=application_name,
            settings=settings,
            station=self.station,
        )
        for interface in interfaces:
            if interface not in self.station.interfaces:
                raise ConfigurationError(
                    "Attempted to register application to non-existent "
                    f"interface, {repr(interface)}.",
                )
            new_address = Address(
                name=station_name,
                ssid=self.next_ssid(
                    interface=self.station.interfaces[interface],
                    name=station_name,
                    ssid=ssid,
                ),
            )
            self._application_map[interface][new_address] = application_instance
            self.register_station(
                interface=self.station.interfaces[interface], address=new_address
            )
        return application_instance

    def register_station(
        self, *, interface: "Interface[Any]", address: Address
    ) -> None:
        """
        Registers an address as existing on a particular interface. It can then be
        looked up later for making connections.
        """
        self._internal_addresses[address] = self._internal_addresses.get(
            address, set()
        ) | {interface}

    def inbound_connect(self, frame: Frame, interface: Interface) -> None:
        """
        Add a connection to the connection table based on an incoming frame.
        """
        first_party = frame.route.src.address
        second_party = frame.route.dest.address
        digipeaters = tuple(
            digipeater.address for digipeater in frame.route.digipeaters
        )
        key = connection_key(first_party, second_party, interface)
        if key in self.table:
            # Existing connection will handle this.
            return
        connection = self.add_connection(
            first_party=frame.route.src.address,
            second_party=frame.route.dest.address,
            interface=interface,
            digipeaters=digipeaters,
            inbound=True,
        )
        connection.negotiate()

    def app_for(
        self,
        interface: Interface,
        destination: Address,
    ) -> BaseApplication[Any]:
        """
        Get an app for a particular address on a specific interface.

        Raises KeyError if no such app exists.
        """
        return self._application_map[interface.name][destination]

    def add_connection(
        self,
        *,
        first_party: Address,
        second_party: Address,
        digipeaters: tuple[Address, ...] = tuple(),
        interface: Interface,
        inbound: bool,
        application: BaseApplication[Any] | None = None,
    ) -> Connection:
        """
        Add a connection to the connection table. Returns a connection if one already
        exists. Note that an existing connection may be outbound rather than inbound,
        or vice versa.
        """
        key = connection_key(first_party, second_party, interface)
        if key in self.table:
            raise ConnectionError(f"Connection already exists! {self.table[key]}")
        if not application and inbound:
            application = self._application_map[interface.name][second_party]
        connection = Connection(
            first_party=first_party,
            second_party=second_party,
            digipeaters=digipeaters,
            interface=interface,
            application=application,
            station=self.station,
            service=self,
            inbound=inbound,
            # Considering outbound connections 'sudo' could result in confusion
            # and potential privilege escalation given a clever loopback trick.
            is_admin=inbound and interface.sudo,
            close_callback=self.connection_closed_callback,
        )
        self.table[key] = connection
        match_call = MatchCall(
            matcher=connection.frame_matcher(),
            notify=connection.inbound_frame,
        )
        self.station.frame_router.register_matcher(
            connection.match_key,
            match_call,
        )
        return connection

    def connection_closed_callback(self, connection: Connection) -> None:
        """
        Callback function to be used when connections are closed out.
        """
        self.station.frame_router.remove_matcher(connection.match_key)
        self.remove_connection(connection)

    def remove_connection(self, connection: Connection) -> None:
        """
        Removes a connection from the connection table.
        """
        key = connection_key(
            connection.first_party, connection.second_party, connection.interface
        )
        del self.table[key]

    def next_ssid(
        self,
        *,
        name: str,
        interface: "Interface[Any]",
        ssid: int | None = None,
    ) -> int:
        """
        Given a station name and an interface, produces the next SSID.

        You can give an ssid you would prefer by specifying the ssid argument. If the
        ssid already exists or is invalid, throws an exception. This function is
        used primarily by the station in order to assign default SSIDs to applications.
        """
        name = name.upper()
        ssids: set[int] = set()
        for address, interfaces in self._internal_addresses.items():
            if address.name != name:
                continue
            if interface in interfaces:
                ssids |= {address.ssid}
        if ssid is None:
            for candidate in range(16):
                if candidate not in ssids:
                    ssid = candidate
                    break
            else:
                raise ConfigurationError(
                    f"All valid SSIDs on {name}, interface {interface} are taken! "
                    "You may have at most 16 SSIDs, numbered 0-15."
                )
        if not 0 <= ssid <= 15:
            raise ConfigurationError("SSID must be between 0 and 15.")
        if ssid in ssids:
            raise ConfigurationError(
                f"SSID {repr(ssid)} already registered on {interface}!"
            )
        return ssid

    async def shutdown(self) -> None:
        """
        Shutdown the connection service, closing out all connections.
        """
        for connection in list(self.table.values()):
            connection.shutdown()



"""
Monitoring service module.

The monitoring service allows the station to keep track of traffic it has heard, and
keep reference tables that can be used. Comes with a couple of commands that you can
add to other applications.
"""

import asyncio
import contextlib
import json
import logging
from ast import literal_eval
from asyncio import Task
from collections.abc import Callable
from dataclasses import dataclass, field
from datetime import UTC, datetime
from pathlib import Path
from typing import TYPE_CHECKING, Literal, Self, cast

from pax25.applications.parsers import autocompleted_enum
from pax25.applications.router import CommandContext, CommandSpec
from pax25.applications.utils import build_table, send_message
from pax25.ax25.constants import UFrameType
from pax25.ax25.control import Unnumbered
from pax25.ax25.frame import Frame
from pax25.ax25.matchers import MatchCall, on_gateway
from pax25.interfaces import Interface
from pax25.protocols import JSONObj
from pax25.services.beacon import next_interval
from pax25.services.connection.connection import Connection, connection_key
from pax25.types import MonitorSettings
from pax25.utils import (
    GatewayDict,
    PortSpec,
    build_json_deserializer,
    cancel_all,
    gateways_for,
    safe_save,
    smart_clone,
)

if TYPE_CHECKING:  # pragma: no cover
    from pax25.station import Station


logger = logging.getLogger(__name__)


@dataclass(kw_only=True)
class HeardEntry:
    """
    Log entry for heard stations.
    """

    last_frame: Frame
    last_heard: datetime = field(default_factory=lambda: datetime.now(UTC))
    interface: str = ""
    last_ui: bytes = b""

    def to_json(self) -> JSONObj:
        return {
            "__class__": self.__class__.__name__,
            "last_frame": self.last_frame.to_json(),
            "last_heard": self.last_heard.isoformat(),
            "interface": self.interface,
            "last_ui": repr(self.last_ui),
        }

    @classmethod
    def from_json(cls, obj: JSONObj) -> Self:
        kwargs = {
            "last_frame": Frame.from_json(cast(JSONObj, obj["last_frame"])),
            "last_heard": datetime.fromisoformat(cast(str, obj["last_heard"])),
            "interface": str(obj["interface"]),
            "last_ui": literal_eval(cast(str, obj["last_ui"])),
        }
        return cls(**kwargs)


@dataclass(kw_only=True, frozen=True)
class LoggedFrame:
    """
    Logged frame heard over the air, with context information.
    """

    frame: Frame
    port: PortSpec


def default_monitoring_settings() -> MonitorSettings:
    """
    Default monitoring settings.
    """
    return {
        "max_frame_log_size": 256,
        "max_stations_tracked": 30,
        "memory_file": "",
        "save_interval": 60,
    }


type LogListener = Callable[[Frame, Interface], bool]

type HeardOptions = Literal["short", "long", "normal"]


class Monitor:
    """
    Service that monitors incoming packets and keeps track of information like what
    stations have been heard and what they've said about themselves.
    """

    def __init__(
        self,
        *,
        station: "Station",
        settings: MonitorSettings | None,
    ):
        """
        Initializes the monitoring service.
        """
        self.station = station
        self._frame_log: list[LoggedFrame] = []
        self._heard_table: dict[str, HeardEntry] = {}
        self._listeners: dict[str, LogListener] = {}
        self._ports_cache = gateways_for(self.station)
        self._settings = default_monitoring_settings()
        self._save_loop: Task[None] | None = None
        if settings:
            self._settings.update(settings)
        self.station.frame_router.register_matcher(
            "monitor", MatchCall(matcher=on_gateway, notify=self.update_entry)
        )
        self.load_stations()

    def refresh_ports_cache(self) -> None:
        """
        Refreshes the ports cache.
        """
        self._ports_cache = gateways_for(self.station)

    @property
    def settings(self) -> MonitorSettings:
        """
        Return the monitor's current settings.
        """
        return smart_clone(self._settings)

    async def reload_settings(self, settings: MonitorSettings) -> None:
        """
        Reload monitoring settings.
        """
        # The settings for monitoring currently don't require any specific action
        # to change.
        await cancel_all([self._save_loop])
        self._settings = settings
        self._save_loop = asyncio.ensure_future(self.save_loop())

    def run(self) -> None:
        """
        Starts up the monitor.
        """
        self._save_loop = asyncio.ensure_future(self.save_loop())

    def load_stations(self) -> None:
        """
        Load heard stations from a save file.
        """
        file_path = self._settings["memory_file"]
        if not file_path:
            return
        object_hook = build_json_deserializer({"HeardEntry": HeardEntry})
        try:
            with open(file_path) as save_file:
                self._heard_table = json.load(save_file, object_hook=object_hook)
        except OSError as err:
            if err.errno == 2:
                # File does not yet exist.
                logger.info(
                    f"No monitoring memory file exists at {repr(file_path)}. "
                    f"We will start a new one.",
                )
            else:
                raise

    async def save_stations(self) -> None:
        """
        Save all the known stations to a file for later retrieval.
        """

        if not self._settings["memory_file"]:
            raise ValueError("No memory file set!")
        save_file_path = Path(self._settings["memory_file"])
        try:
            safe_save(path=save_file_path, data=self._heard_table, debug=False)
        except Exception as err:
            logger.exception(err)

    async def save_loop(self) -> None:
        """
        Loop that periodically saves all the stations we've heard.
        """
        if self._settings["memory_file"] is None:
            return
        next_timestamp = datetime.now(UTC)
        while True:
            next_timestamp = await next_interval(
                next_timestamp, self._settings["save_interval"]
            )
            await self.save_stations()

    def update_entry(self, frame: Frame, interface: Interface) -> None:
        """
        Updates the entry in the heard table.
        """
        source = str(frame.route.src.address)
        entry = self._heard_table.get(source) or HeardEntry(
            last_frame=frame,
        )
        entry.last_heard = datetime.now(UTC)
        entry.interface = interface.name
        if isinstance(frame.control, Unnumbered) and (
            frame.control.frame_type == UFrameType.UNNUMBERED_INFORMATION
        ):
            entry.last_ui = frame.info
        self._heard_table[source] = entry
        self.log_frame(frame, interface)
        max_tracked = self._settings["max_stations_tracked"]
        if (max_tracked is not None) and (len(self._heard_table) > max_tracked):
            new_table: dict[str, HeardEntry] = {}
            revised_set = list(
                sorted(self._heard_table.items(), key=lambda x: x[1].last_heard)
            )[-max_tracked:]
            for address, entry in revised_set:
                new_table[address] = entry
            self._heard_table = new_table

    def log_frame(self, frame: Frame, interface: Interface) -> None:
        """
        Log a frame heard on a gateway.
        """
        route_key = connection_key(
            frame.route.src.address, frame.route.dest.address, interface
        )
        if route_key in self.station.connection.table:
            # We don't log frames we're active party to. This may change in the future,
            # delegating filtering downstream.
            return
        # If any of our frame listeners return True, we don't log. We consider the frame
        # 'consumed' from the log in that case.
        if any([listener(frame, interface) for listener in self._listeners.values()]):
            return
        self._frame_log.append(
            LoggedFrame(frame=frame, port=self._ports_cache[interface])
        )
        if self._settings["max_frame_log_size"] is None:
            return
        if self._settings["max_frame_log_size"] < len(self._frame_log):
            del self._frame_log[: -self._settings["max_frame_log_size"]]

    def _build_std_table(
        self,
        heard_stations: dict[str, HeardEntry],
        gateways: GatewayDict,
        include_dest: bool = False,
    ) -> list[str]:
        lines = []
        for address, entry in heard_stations.items():
            try:
                port = gateways[self.station.interfaces[entry.interface]].number
            except KeyError:
                # No longer a gateway, or the interface is down.
                continue
            marker = ""
            if entry.last_frame.route.digipeaters:
                marker = "*"
            line = [f"{port}:{address}{marker}"]
            if include_dest:
                line.append(f"> {entry.last_frame.route.dest}")
            last_heard = entry.last_heard.astimezone()
            line.append(str(last_heard.date()))
            line.append(str(last_heard.time()).split(".")[0])
            lines.append(line)
        return build_table(lines)

    def _std_table(self, connection: Connection, gateways: GatewayDict) -> None:
        """
        Print a standard heard table.
        """
        heard_stations = self.heard_stations
        table = self._build_std_table(heard_stations, gateways)
        send_message(connection, "\r".join(table))

    def _short_table(self, connection: Connection, gateways: GatewayDict) -> None:
        """
        Print a short heard table.
        """
        lines = []
        for address, entry in self.heard_stations.items():
            try:
                port = gateways[self.station.interfaces[entry.interface]].number
            except KeyError:
                # No longer a gateway, or the interface is down.
                continue
            marker = ""
            if entry.last_frame.route.digipeaters:
                marker = "*"
            line = f"{port}:{address}{marker}"
            lines.append(line)
        send_message(connection, "\r".join(lines))

    def _long_table(self, connection: Connection, gateways: GatewayDict) -> None:
        """
        Print a long heard table.
        """
        heard_stations = self.heard_stations
        table = self._build_std_table(heard_stations, gateways, include_dest=True)
        final_lines = []
        for summary_line, entry in zip(table, heard_stations.values(), strict=False):
            final_lines.append(summary_line)
            if entry.last_frame.route.digipeaters:
                final_lines.append(
                    " VIA "
                    + ",".join(
                        str(digipeater)
                        for digipeater in entry.last_frame.route.digipeaters
                    )
                )
            last_ui = entry.last_ui.decode("utf-8", errors="ignore").strip()
            if last_ui:
                final_lines.append(" UI: " + last_ui)
        send_message(connection, "\r".join(final_lines))

    def heard(
        self, connection: Connection, context: CommandContext[HeardOptions]
    ) -> None:
        """
        Heard command. Can be included using the 'command' property on the monitoring
        service, which defines a CommandSpec for inclusion in other apps.
        """
        gateways = gateways_for(self.station)
        match context.args:
            case "short":
                self._short_table(connection, gateways)
            case "long":
                self._long_table(connection, gateways)
            case "normal":
                self._std_table(connection, gateways)

    def add_listener(self, label: str, listener: LogListener) -> None:
        """
        Registers a frame listener.
        """
        if label in self._listeners:
            logger.warning(
                f"Duplicate label {repr(label)} registered as monitoring listener. "
                f"This is likely a bug or configuration issue.",
            )
        self._listeners[label] = listener

    def remove_listener(self, label: str) -> None:
        """
        Removes a frame listener.
        """
        if label not in self._listeners:
            logger.warning(
                f"Removing non-existent monitoring listener, {repr(label)}. This is "
                f"likely a bug or configuration issue.",
            )
            return
        del self._listeners[label]

    def consume_logs(self) -> list[LoggedFrame]:
        """
        Returns the current frame log, emptying it.
        """
        log = self._frame_log
        self._frame_log = []
        return log

    async def shutdown(self) -> None:
        """
        Shut down the monitoring service.
        """
        await cancel_all([self._save_loop])
        self._save_loop = None
        # One last save for good measure.
        with contextlib.suppress(ValueError):
            await self.save_stations()

    @property
    def heard_stations(self) -> dict[str, HeardEntry]:
        """
        Return a tuple of heard stations.
        """
        return {
            key: value
            for key, value in sorted(
                self._heard_table.items(), key=lambda x: x[1].last_heard
            )
        }

    @property
    def command(self) -> CommandSpec[HeardOptions]:
        """
        Builds a command spec for the heard command.
        """
        return CommandSpec(
            command="heard",
            aliases=("j",),
            help="List all stations heard by this station. Use 'heard long' for "
            "extended info, or 'heard short' for abbreviated info.",
            function=self.heard,
            parser=autocompleted_enum(
                ("short", "long", "normal"),
                default="normal",
            ),
        )



"""
Main application module of the contributed bulletin board system.
"""

import json
from collections.abc import Iterable
from dataclasses import dataclass
from datetime import UTC, datetime
from pathlib import Path
from threading import Lock
from typing import Literal, cast

from pax25 import Application
from pax25.applications.help import Help
from pax25.applications.parsers import (
    ParseError,
    address_parser,
    integer_parser,
    no_arguments,
    pull_segment,
    raw_string_parser,
    string_parser,
)
from pax25.applications.router import (
    CommandContext,
    CommandRouter,
    CommandSpec,
    ParserSpec,
)
from pax25.applications.utils import send_message
from pax25.ax25.address import Address
from pax25.contrib.bulletin_board.types import (
    BoardConnections,
    BoardSettings,
    BoardState,
    Message,
    ReaderState,
)
from pax25.services.connection.connection import Connection
from pax25.types import Version
from pax25.utils import safe_save, version_string

DEFAULT_MAX_LENGTH = 5000
DEFAULT_SLOTS = 100


def message_headers(message: Message) -> str:
    """
    Format the main headers of a message.
    """
    created_on = (
        str(datetime.fromisoformat(message["created_on"]).strftime("%d/%m/%Y %H:%M:%S"))
        + " "
    )
    to_callsign = f"{message['to_callsign']}".ljust(7, " ")
    from_callsign = f"{message['from_callsign']}".ljust(7, " ")
    return f"{created_on}{to_callsign}{from_callsign}"


def is_party(name: str, message: Message) -> bool:
    """
    Checks if a name (callsign) is party to a message.
    """
    return any(
        (
            message["from_callsign"] == name,
            message["to_callsign"] == name,
        )
    )


def message_preview(message: Message) -> str:
    """
    Create a message preview string, like what would be used in message listings.
    """
    id_text = f"{message['id']}".ljust(4, " ")
    status_text = "R" if message["read"] else "U"
    status_text += "!" if message["private"] else ""
    status_text = status_text.ljust(6, " ")
    message_size = len(message["body"].encode("utf-8")) + len(
        message["subject"].encode("utf-8")
    )
    size = str(f"{message_size}").ljust(5, " ")
    headers = message_headers(message)
    subject_snippet = message["subject"][:30]
    return f"{id_text}{status_text}{size}{headers}{subject_snippet}"


@dataclass
class ListOpts:
    """
    Arguments for the list command.
    """

    direction: Literal[">", "<"] | None
    callsign: Address | None


def listing_parser(spec: ParserSpec) -> ListOpts:
    """
    Parser for the list command.
    """
    args = spec.args.strip()
    if not args:
        return ListOpts(direction=None, callsign=None)
    direction: Literal[">", "<"] | None = None
    direction_segment, remainder = pull_segment(spec.args)
    if direction_segment and direction_segment not in [">", "<"]:
        raise ParseError(
            f"Unrecognized direction: {repr(direction_segment)}. Options are > or <."
        )
    elif direction_segment:
        direction = cast(Literal[">", "<"], direction_segment)
    address = remainder.strip()
    try:
        callsign = Address.from_string(address)
    except ValueError as err:
        raise ParseError(f"{repr(address)} is not a valid callsign.") from err
    return ListOpts(
        direction=direction,
        callsign=callsign,
    )


PROMPT_TEXT = "B,L,K,R,S, H(elp) or I(nfo) >"


class BulletinBoard(Application[BoardSettings]):
    """
    A Personal Bulletin Board System (PBBS) application.

    The bulletin board has the ability to persist its data to disk. By default,
    it will store its data in the current working directory, but the storage file can
    be customized.
    """

    version = Version(major=0, minor=2, patch=0)
    version_string = f"[PAX25-PBBS-{version_string(version)}]"
    short_name = "B"
    board_state: BoardState
    connections: BoardConnections
    _next_id = 1
    _save_lock: Lock
    routers: dict[Literal["home", "subject", "body"], CommandRouter]

    def setup(self) -> None:
        """
        Set up the basic state of the bulletin board system.
        """
        self.connections = {}
        self._save_lock = Lock()
        self.board_state = self.load_board_state()
        self.routers = {
            "home": CommandRouter(
                post_command_func=self.send_prompt,
                default=CommandSpec(
                    command="",
                    help="",
                    parser=string_parser,
                    function=lambda x, _y: self.send_prompt(x),
                ),
            ),
            "subject": CommandRouter(
                default=CommandSpec(
                    command="",
                    help="",
                    parser=string_parser,
                    function=self.set_subject,
                )
            ),
            "body": CommandRouter(
                post_command_func=self.send_prompt,
                default=CommandSpec(
                    command="",
                    help="",
                    parser=raw_string_parser,
                    function=self.handle_body_line,
                ),
            ),
        }
        help_command = Help(self.routers["home"])
        self.routers["home"].add(
            CommandSpec(
                command="list",
                help="\r".join(
                    [
                        "L(ist)       List messages you can read",
                        "L <|> call   List messages to or from a callsign",
                    ]
                ),
                aliases=("l",),
                parser=listing_parser,
                function=self.list_messages,
            ),
            CommandSpec(
                command="lmine",
                help="LM(ine)\rList unread messages addressed to you",
                parser=no_arguments,
                function=self.list_mine,
            ),
            CommandSpec(
                command="read",
                help="R(ead) n\rDisplay message ID n",
                aliases=("r",),
                parser=integer_parser,
                function=self.read_message,
            ),
            CommandSpec(
                command="rmine",
                help="RM(ine)\rRead all unread messages addressed to you",
                parser=no_arguments,
                function=self.read_mine,
            ),
            CommandSpec(
                command="send",
                help="S(end) call\rSend message to callsign",
                parser=address_parser,
                aliases=("s",),
                function=self.compose_message,
            ),
            CommandSpec(
                command="sprivate",
                help="SP(rivate) call\rSend a private message to a callsign.",
                parser=address_parser,
                function=self.send_private,
            ),
            CommandSpec(
                command="kill",
                help="Deletes a message.",
                aliases=("k",),
                parser=integer_parser,
                function=self.kill_message,
            ),
            CommandSpec(
                command="kmine",
                help="Deletes all messages from and to you.",
                parser=no_arguments,
                function=self.kill_mine,
            ),
            CommandSpec(
                command="bye",
                help="B(ye)\rBBS will disconnect",
                function=self.bye,
                parser=no_arguments,
                aliases=("quit",),
            ),
            CommandSpec(
                command="info",
                help="I(nfo)\rLearn about his PBBS software",
                function=self.send_info,
                parser=no_arguments,
            ),
            help_command.spec,
        )

    @property
    def save_file_path(self) -> Path:
        """Get the path to the board's save file."""
        return Path(self.settings.get("save_file_path", "board.json"))

    @property
    def welcome_message(self) -> str:
        """Get the welcome message for the board."""
        welcome_message = self.settings.get("welcome_message") or ""
        if welcome_message:
            welcome_message += "\r"
        return welcome_message

    def load_board_state(self) -> BoardState:
        """
        Load the board's state from the file path specified in the settings.
        """
        if not self.save_file_path.is_file():
            return BoardState(messages={}, version=self.version)
        with open(self.save_file_path, encoding="utf-8") as save_file:
            struct = json.load(save_file)
            loaded_state = cast(BoardState, struct)
            return loaded_state

    def send_prompt(self, connection: Connection) -> None:
        """
        Send the command prompt.
        """
        if connection not in self.connections:
            return
        state = self.connections[connection]
        if state.mode == "home":
            send_message(connection, "B,L,K,R,S, H(elp) or I(nfo) >")

    @property
    def next_id(self) -> int:
        """
        Get the next available board message ID. We recycle board IDs since
        we could be running for a very long time, and we have space to burn.
        """
        if str(self._next_id) not in self.board_state["messages"]:
            return self._next_id
        while str(self._next_id) in self.board_state["messages"]:
            self._next_id += 1
        return self._next_id

    def _save_board_state(self) -> None:
        """
        Persist the board's state to disk. First write to a temporary file, then
        overwrite the original. This way we can revert to backup if we fail mid-dump.

        Don't call this directly. .save() instead.
        """
        safe_save(
            path=self.save_file_path,
            data=self.board_state,
            debug=bool(self.settings.get("debug")),
        )

    def save(self) -> None:
        """
        Locks the db for save and saves it in a thread-safe manner.
        """
        with self._save_lock:
            self._save_board_state()

    def reset_state(self, connection: Connection) -> None:
        """
        Set the user's state to a blank default-- puts them to the home mode.
        """
        self.connections[connection] = ReaderState(
            mode="home",
            body="",
            subject="",
            to_callsign="",
            private=False,
        )

    def on_startup(self, connection: Connection) -> None:
        """Set up the user's connection state and greet them."""
        self.reset_state(connection)
        callsign = connection.first_party.name
        unread_count = len(
            [
                message
                for message in self.board_state["messages"].values()
                if message["to_callsign"] == callsign and not message["read"]
            ]
        )
        unread_segment = ""
        if unread_count:
            plural = "s" if unread_count != 1 else ""
            unread_segment = f"\rYou have {unread_count} unread message{plural}."
        slots_status = ""
        slots = self.settings.get("slots", DEFAULT_SLOTS)
        if slots is not None:
            remaining = max(0, slots - len(self.board_state["messages"]))
            plural = "s" if slots != 1 else ""
            slots_status = f"\r{remaining} of {slots} message slot{plural} available."
        greeting = (
            f"{self.version_string}{self.welcome_message}{unread_segment}{slots_status}"
        )
        greeting += f"\r{PROMPT_TEXT}"
        send_message(connection, greeting)

    def on_shutdown(self, connection: Connection) -> None:
        """Clear the user's connection state."""
        del self.connections[connection]

    def handle_body_line(
        self,
        connection: Connection,
        context: CommandContext[str],
    ) -> None:
        """Handle input when composing a message body."""
        # Default maximum message length is 5000 characters.
        state = self.connections[connection]
        size_limit = self.settings.get("max_message_length", DEFAULT_MAX_LENGTH)
        message = context.raw_input
        if message.rstrip() != "/EX":
            state.body += message + "\r"
            if size_limit is not None:
                state.body = state.body[:size_limit]
                if len(state.body) >= size_limit:
                    send_message(
                        connection,
                        "Message body size limit reached. Type /EX to exit.",
                    )
                    return
            return
        self.board_state["messages"][str(self.next_id)] = Message(
            id=self.next_id,
            to_callsign=state.to_callsign,
            from_callsign=connection.first_party.name,
            subject=state.subject,
            body=state.body,
            created_on=datetime.now(UTC).astimezone().isoformat(),
            read=False,
            private=state.private,
        )
        self.save()
        self.reset_state(connection)
        send_message(connection, "Message saved.")

    def send_listing(self, connection: Connection, messages: Iterable[Message]) -> None:
        """
        Given a list of messages, send all of them to the client.
        """
        # May be a generator. Resolve, in that case, so we can check for emptiness.
        messages = list(messages)
        if not messages:
            send_message(connection, "No messages available.")
            return
        send_message(
            connection,
            "ID# Flags Size Date        Time     To     From   Subject",
        )
        send_message(
            connection,
            "\r".join([message_preview(message) for message in messages]),
        )

    def list_mine(self, connection: Connection, _context: CommandContext[None]) -> None:
        """
        List all messages addressed to the current user.
        """
        self.send_listing(
            connection,
            [
                message
                for message in reversed(self.board_state["messages"].values())
                if message["to_callsign"] == connection.first_party.name
            ],
        )

    def can_read(self, connection: Connection, message: Message) -> bool:
        """
        Determines if a connected user has the right to read this message.
        """
        if connection.is_admin:
            return True
        if not message["private"]:
            return True
        return is_party(connection.first_party.name, message)

    def readable(self, connection: Connection) -> Iterable[Message]:
        """
        Yields all messages which can be read by a connected user.
        """
        for message in reversed(self.board_state["messages"].values()):
            if self.can_read(connection, message):
                yield message

    def list_messages(
        self,
        connection: Connection,
        context: CommandContext[ListOpts],
    ) -> None:
        """
        List messages.
        """
        if not (context.args.callsign and context.args.direction):
            self.send_listing(
                connection,
                self.readable(connection),
            )
            return

        attr: Literal["from_callsign", "to_callsign"]
        attr = "from_callsign" if context.args.direction == "<" else "to_callsign"
        self.send_listing(
            connection,
            [
                message
                for message in self.readable(connection)
                if message[attr] == str(context.args.callsign.name)
            ],
        )

    def read_mine(self, connection: Connection, _context: CommandContext[None]) -> None:
        """
        Read all messages addressed to the current user.
        """
        messages = [
            message
            for message in reversed(self.board_state["messages"].values())
            if message["to_callsign"] == connection.first_party.name
            and not message["read"]
        ]
        for message in messages:
            self.perform_read(connection, message, save=False)
        self.save()

    def perform_read(
        self, connection: Connection, message: Message, save: bool = True
    ) -> None:
        """
        Send a message to a user, and mark it read.
        """
        private_prefix = "!!PRIVATE!! " if message["private"] else ""
        body = message["body"]
        # Should always have a newline at the end.
        if body and body[-1] != "\r":
            body += "\r"
        send_message(
            connection,
            f"ID#{message['id']} {message_headers(message)}\r{private_prefix}SUBJECT: "
            f"{message['subject']}\r{body}",
            append_newline=False,
        )
        if message["to_callsign"] == connection.first_party.name:
            message["read"] = True
            if save:
                self.save()

    def read_message(
        self,
        connection: Connection,
        context: CommandContext[int],
    ) -> None:
        """
        Read a given message ID
        """
        number = str(context.args)
        if (number not in self.board_state["messages"]) or not (
            self.can_read(connection, self.board_state["messages"][number])
        ):
            send_message(connection, f"Could not find message with ID {number}")
        message = self.board_state["messages"][number]
        self.perform_read(connection, message)

    def bye(self, connection: Connection, _context: CommandContext[None]) -> None:
        """
        Command for closing the connection.
        """
        send_message(connection, "Goodbye!")
        connection.disconnect()

    def compose_message(
        self,
        connection: Connection,
        context: CommandContext[Address],
        private: bool = False,
    ) -> None:
        """
        Compose a message. We'd call this function 'send' or 'send_message', but those
        are reserved by the parent class.

        More specifically, this starts the composition by changing the mode. The rest
        of the composition work is handled in on_message.
        """
        slots = self.settings.get("slots", DEFAULT_SLOTS)
        if slots is not None and slots <= len(self.board_state["messages"]):
            send_message(
                connection,
                "The message board is full. You cannot send any more messages.",
            )
            return
        state = self.connections[connection]
        state.to_callsign = context.args.name
        state.mode = "subject"
        state.private = private
        send_message(connection, "SUBJECT: ", append_newline=False)

    def send_private(
        self, connection: Connection, context: CommandContext[Address]
    ) -> None:
        """
        Send a private message.
        """
        self.compose_message(connection, context, private=True)

    def kill_mine(self, connection: Connection, _context: CommandContext[None]) -> None:
        """
        Kills all read messages addressed to the current user.
        """
        count = 0
        lowest = self._next_id
        name = connection.first_party.name
        # Iterating over a coerced list here so that we're not modifying the dictionary
        # as the key, value pairs are being generated.
        for key, message in list(self.board_state["messages"].items()):
            if message["read"] and message["to_callsign"] == name:
                lowest = min(lowest, message["id"])
                del self.board_state["messages"][key]
                count += 1
        if count:
            self.save()
        self._next_id = lowest
        plural = "" if count == 1 else "s"
        send_message(connection, f"\r{count} message{plural} deleted.")

    def kill_message(
        self,
        connection: Connection,
        context: CommandContext[int],
    ) -> None:
        """
        Delete a message from the database.
        """
        number = str(context.args)
        if number not in self.board_state["messages"]:
            send_message(
                connection,
                f"Could not find message with ID {number}.",
            )
        message = self.board_state["messages"][number]
        if connection.is_admin or is_party(connection.first_party.name, message):
            del self.board_state["messages"][number]
            self.save()
            self._next_id = min([self._next_id, int(number)])
            send_message(connection, f"Message {number} deleted.")
            return
        # Do not reveal this message exists.
        if message["private"]:
            send_message(connection, f"Could not find message with ID {number}.")
        send_message(
            connection,
            f"You do not have permission to kill message {number}.",
        )
        return

    def send_info(self, connection: Connection, _context: CommandContext[None]) -> None:
        """
        Send an informational message about this board system.
        """
        info_lines = [
            "## Pax25 PBBS",
            "Welcome to the Pax25 reference bulletin board implementation. Pax25 is a "
            "python library for creating packet radio applications. You are invited to "
            "join in on the fun by visiting us at: ",
            "",
            "https://foxyfoxie.gitlab.io/pax25/",
            "",
            "Pax25 was developed by KW6FOX, K1LEO, and KF0KAA. Additional contribution "
            "credit can be found on the GitLab repository homepage.",
        ]
        send_message(connection, "\r".join(info_lines))

    def set_subject(self, connection: Connection, spec: CommandContext[str]) -> None:
        """
        Sets the subject of the mail.
        """
        state = self.connections[connection]
        state.subject = spec.raw_input[:150]
        state.mode = "body"
        response = "ENTER MESSAGE--END WITH /EX ON A SINGLE LINE"
        max_length = self.settings.get("max_message_length", DEFAULT_MAX_LENGTH)
        if max_length is not None:
            response += f" (MAX {max_length} CHARS)"
        send_message(connection, response)

    def on_message(self, connection: Connection, message: str) -> None:
        """
        Perform the command routing.
        """
        state = self.connections[connection]
        self.routers[state.mode].route(connection, message)



"""
Protocols for duck typing elements of AX.25 serialization/deserialization.

Do not include any code that should actually run here, as it won't be analyzed for
test coverage.
"""

from typing import Protocol, Self


class Assembler(Protocol):  # pragma: no cover
    """
    Protocol for logical segments of a packet. Allows for assembly/disassembly.

    Also includes __len__ for Sized compatibility because you cannot intersect
    protocols with MyPy (yet.)
    """

    def size(self) -> int:
        """
        The length of the data structure when assembled.
        """
        raise NotImplementedError("Subclasses must implement size.")

    def assemble(self) -> bytes:
        """
        Method for turning a data structure into bytes suitable for transmission.
        """

    @classmethod
    def disassemble(cls, data: bytes) -> Self:
        """
        Method for instantiating this data structure from bytes.
        """



"""
Utility functions/coroutines for the pax25 project.
"""

import asyncio
import json
import os
import random
import string
from asyncio import CancelledError, Future, Task
from collections import defaultdict
from collections.abc import Callable, Coroutine, Generator, Iterable
from contextlib import suppress
from dataclasses import dataclass
from functools import partial, singledispatch
from json import JSONEncoder
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    ParamSpec,
    TypeVar,
    cast,
)

from pax25.protocols import JSONObj, JSONSerializable

if TYPE_CHECKING:  # pragma: no cover
    from pax25.interfaces.types import Interface
    from pax25.station import Station
    from pax25.types import Updatable, Version

P = ParamSpec("P")
R = TypeVar("R")


def async_wrap(func: Callable[P, R]) -> Callable[P, Coroutine[None, None, R]]:
    """
    Wraps a function that requires syncronous operation so it can be awaited instead
    of blocking the thread.

    Shamelessly stolen and modified from:
    https://dev.to/0xbf/turn-sync-function-to-async-python-tips-58nn
    """

    async def run(*args: P.args, **kwargs: P.kwargs) -> R:
        loop = asyncio.get_event_loop()
        part = partial(func, *args, **kwargs)
        return await loop.run_in_executor(None, part)

    return run


class EnumReprMixin:
    """
    Mixin for enums that allows their export to remain copy-pastable for instantiation.
    """

    def __repr__(self) -> str:
        """
        Repr for an Enum that allows for copy-pastable instantiation.
        """
        return f"{self.__class__.__name__}.{self._name_}"  # type: ignore [attr-defined]


def digipeater_factor(*, interval: int, hops: int, constant: int = 1000) -> int:
    """
    Returns a modified length of time based on the number of digipeaters used.

    Each digipeater increases the amount of time required between retries.
    """
    # For now, assuming this is a straight multiplication factor, though maybe we should
    # add some additional padding for processing time.
    constant = constant if hops else 0
    return (hops + 1) * interval + constant


async def cancel(task: Task[Any]) -> None:
    """
    Cancels a task, and then waits for it to be cleaned up.
    """
    task.cancel()
    with suppress(CancelledError):
        await task


async def cancel_all(tasks: Iterable[Task[Any] | None]) -> list[None]:
    """
    Cancel a series of tasks. Returns a None for each task as they are cancelled,
    useful for closing out optionally running tasks.
    """
    results: list[None] = []
    for task in tasks:
        if task is not None:
            await cancel(task)
        results.append(None)
    return results


@singledispatch
def normalize_line_endings(data: bytes) -> bytes:
    """
    Normalize line endings in a bytestring to be sensible on the target sytem.
    """
    return (
        data.replace(b"\r\n", b"\r")
        .replace(b"\n", b"\r")
        .replace(b"\r", os.linesep.encode("utf-8"))
    )


def generate_nones(times: int | None) -> Generator[None]:
    """
    Generator that yields however many Nones as you specify, or an infinite number
    of Nones if times is None.

    This is primarily to normalize loops that could be either a certain amount of
    iterations or forever.
    """
    if times is None:
        while True:
            yield
    else:
        for _ in range(times):
            yield


@dataclass(kw_only=True, frozen=True)
class PortSpec:
    """
    Named tuple for Interface identities as enumerated by the gateways_for function.
    """

    number: int
    name: str
    type: str


type GatewayDict = dict["Interface", PortSpec]


def gateways_for(station: "Station") -> GatewayDict:
    """
    Retrieve all gateways from a station in numerical order.
    """
    gateways: dict[Interface, PortSpec] = {}
    number = 0
    for key, value in station.interfaces.items():
        if value.gateway:
            number += 1
            gateways[value] = PortSpec(number=number, name=key, type=value.type)
    return gateways


def smart_clone[T](
    item: T,
) -> T:
    """
    This makes a copy of item, with a few caveats:

    1. The copy will copy all references to primitive immutable items (str, int, float).
    2. The copy will create independent copies of all containers it understands.
    3. If the object is not recognized and mutable, copies reference.
    4. It can't handle self-referential structures. So don't hand it any, or you'll get
       an infinite loop.

    The standard clone function is a shallow clone that doesn't clone recursively.
    The deep cloning function is TRULY deep in a way that could cause problems in the
    case that we're cloning something like the settings of the file interface, and it's
    been handed a buffer.

    This is a middle ground-- the primary data structures are all rebuilt so their
    contents aren't affected when you modify the result you get. However, for objects
    which are more complicated than primitives and the basic data structures, we copy
    existing references rather than cloning deeply.
    """
    result: Any

    match item:
        case tuple():
            # Could be a NamedTuple, so clone with class.
            result = item.__class__(*(smart_clone(entry) for entry in item))
        case list():
            result = [smart_clone(entry) for entry in item]
        case defaultdict():
            result = defaultdict(item.default_factory)
            for key, value in item.items():
                result[smart_clone(key)] = smart_clone(value)
        case dict():
            result = {
                smart_clone(key): smart_clone(value) for key, value in item.items()
            }
        case set():
            result = {smart_clone(item) for item in item}
        case _:
            result = item
    return cast(T, result)


D = TypeVar("D", bound="Updatable")


def maybe_update(to_update: D, new_values: None | D) -> D:
    """
    Updates a dictionary if new_values is not None. Useful for config structs which
    may not have a value, to avoid needlessly verbose if chains.

    Returns the updated dict for convenience, but it is indeed mutated in place.
    """
    if new_values is None:
        return to_update
    to_update.update(new_values)
    return to_update


def random_ascii_data() -> bytes:
    """
    Generates a significant amount of ASCII bytes data for transfer-- enough to exhaust
    the frame window.
    """
    data = bytearray()
    letters = string.ascii_lowercase
    for _ in range(10):
        # 256 is the default MTU, and we want our lines to be longer than that so they
        # have to be broken up. We also don't want the commands to be exactly the size
        # of the MTU so we know they're sent/received correctly.
        data.extend(
            int.from_bytes(random.choice(letters).encode("ascii")) for __ in range(300)
        )
        data.append(int.from_bytes(b"\r"))
    return bytes(data)


async def first(*args: Task[None] | Future[None]) -> None:
    """
    Waits until the first in a set of futures returns.
    """
    await asyncio.wait(
        tuple(task for task in args if task),
        return_when=asyncio.FIRST_COMPLETED,
    )


def version_string(version: "Version") -> str:
    """
    Constructs a version string from a version dictionary.
    """
    return f"{version['major']}.{version['minor']}.{version['patch']}"


class LazyRepr:
    """
    Utility class that stringifies to the repr of the object handed to it. Useful to
    make sure the logs show the full details of a packet without going through the
    translation even when we're not debugging.
    """

    def __init__(self, target: Any):
        self.target = target

    def __str__(self) -> str:
        return repr(self.target)

    def __eq__(self, other: Any) -> bool:
        """
        Should allow us to test for this kind of argument in tests.
        """
        if not isinstance(other, LazyRepr):
            return False
        return bool(other.target == self.target)


class Pax25JSONEncoder(JSONEncoder):
    """
    Custom encoder for heard entries.
    """

    def default(self, obj: Any) -> JSONObj:
        if hasattr(obj, "to_json"):
            return cast(JSONObj, obj.to_json())
        return cast(JSONObj, super().default(obj))


def build_json_deserializer(
    class_dict: dict[str, type[JSONSerializable]],
) -> Callable[[JSONObj], JSONObj | JSONSerializable]:
    """
    Builds a JSON deserializer with special recognition of particular classes.
    """

    def json_deserializer(obj: JSONObj) -> JSONSerializable | JSONObj:
        class_name = obj.get("__class__", None)
        if class_name is None:
            return obj
        if class_name not in class_dict:
            return obj
        return class_dict[str(class_name)].from_json(obj)

    return json_deserializer


def safe_save(*, path: Path, data: Any, debug: bool = False) -> None:
    """
    Saves to a specific path, using an intermediary temporary file. This prevents
    situations where saving the file results in an unusable state should failure happen
    mid-write. Raises IOError if there's a failure.
    """
    cookie = "".join(random.choice(string.ascii_lowercase) for _ in range(10))
    tmp_path = path.with_suffix(f".{cookie}.tmp")
    try:
        with open(tmp_path, "w", encoding="utf-8") as config_file:
            json.dump(
                data,
                config_file,
                indent=2 if debug else None,
                cls=Pax25JSONEncoder,
            )
    except Exception as err:
        raise OSError(
            f"Failed when saving to temporary file, {repr(tmp_path)}. "
            f"{err.__class__}: {err}"
        ) from err
    os.replace(tmp_path, path)



from typing import Protocol, Self

type JSON = dict[str, "JSON"] | list["JSON"] | str | int | float | bool | None
type JSONObj = dict[str, "JSON"]


class JSONSerializable(Protocol):
    """
    Protocol for objects which can be serialized/deserialized via json.
    """

    def to_json(self) -> JSONObj: ...

    @classmethod
    def from_json(cls, data: JSONObj) -> Self: ...



"""
Exception classes.
"""


class ConfigurationError(Exception):
    """
    To be raised when the client developer has made an error in configuring their
    station or any of its relevant components.
    """



"""
Timer functionality. Used for setting up timeout callbacks which can be cancelled as
needed.
"""

import asyncio
from asyncio import CancelledError, Task
from collections.abc import Callable
from enum import Enum
from logging import getLogger

from pax25.utils import EnumReprMixin, generate_nones

log = getLogger(__name__)


class TimerState(EnumReprMixin, Enum):
    """Enum for timer status."""

    INACTIVE = "inactive"
    RUNNING = "running"


class Timer:
    """
    Timer class. Keeps track of outstanding time, an action to perform, and if it's been
    cancelled.
    """

    def __init__(self, name: str) -> None:
        """
        Set up timer. Does not activate. For that, use the schedule function.
        """
        self.name = name
        self.killed = False
        self._sleep_task: Task[None] | None = None

    def __str__(self) -> str:
        """
        String representation of this timer.
        """
        return f"Timer ({self.name} [{self.state.value}])"

    @property
    def state(self) -> TimerState:
        """
        Return the state of the current timer.
        """
        return TimerState.RUNNING if self._sleep_task else TimerState.INACTIVE

    async def run(self, delay: int) -> bool:
        """
        Start this timer, and wait until the time is complete or the timer is cancelled.

        If the task is cancelled, return True. If the timeout elapsed, return False. The
        reason we return true is so you can write something like:

        ```
        if await timer.run(100):
            return
        # Do something if the timer expired here.
        ```

        See the Station class for examples in the `retry_loop` and `delay_action`
        methods.

        If the timer is already running, cancels the old task to replace with the new
        timer.
        """
        if self.killed:
            # Killed timers act as though their tasks have always been cancelled.
            return True
        if self._sleep_task:
            self.cancel()
        task = asyncio.ensure_future(asyncio.sleep(0.001 * delay))
        try:
            self._sleep_task = task
            await self._sleep_task
        except CancelledError:
            return True
        return False

    def cancel(self) -> None:
        """
        Cancels a timer if it's running. Otherwise, does nothing.
        """
        if self._sleep_task is None:
            return
        self._sleep_task.cancel("Timer cancelled.")
        self._sleep_task = None

    def kill(self) -> None:
        """
        Kills a timer. This means that the timer will no longer accept new tasks.
        """
        self.cancel()
        self.killed = True


def retry_loop(
    *,
    timer: Timer,
    retries: int,
    interval: int,
    immediate: bool = True,
    check: Callable[[], bool] = lambda: False,
    action: Callable[[], None],
    fail_action: Callable[[], None] = lambda: None,
) -> Task[None]:
    """
    Retries an action multiple times at a given interval.
    The action will always be run at least once if immediate is true.
    The number of retries determines how many attempts will be made after this.

    `action` is a callable to perform.
    `check` is a callable that will return True if the action we're calling was
    successful, or `false` if not. If not specified, check is a lambda that always
    returns False.
    `fail_action` will be called if all retries failed. If no function is specified,
    performs a no-op.
    `interval` is the number of milliseconds between attempts.

    This function will exit early if the timer it is using was cancelled by an
    outside action. In this case, the fail_action will not be run.
    """
    # Basic sanity checks to avoid infinite loops.
    if retries is not None:
        assert retries >= 0
    assert interval >= 1

    # Cancel any outstanding items before beginning.
    timer.cancel()

    async def wrapped_retry() -> None:
        if immediate:
            action()
        for _ in generate_nones(retries):
            if await timer.run(interval):
                # Timer was cancelled. We bail.
                return
            if check():
                # Check succeeded. We bail.
                return
            action()
        if await timer.run(interval):
            return
        if check():
            return
        fail_action()

    future = asyncio.ensure_future(wrapped_retry())
    return future


def delay_action(*, timer: Timer, action: Callable[[], None], delay: int) -> Task[None]:
    """
    Performs an action if the timer elapses with the given delay in milliseconds.
    """
    assert delay >= 1
    timer.cancel()

    async def wrapped_delayed_action() -> None:
        if await timer.run(delay):
            return
        action()

    return asyncio.ensure_future(wrapped_delayed_action())



"""
Module for the Application class, used for creating all applications.
"""

import sys
from typing import TYPE_CHECKING

from .types import BaseApplication, BasicStateTable, S

if TYPE_CHECKING:  # pragma: no cover
    from pax25.services.connection.connection import Connection
    from pax25.station import Station


class Application(BaseApplication[S]):
    """
    The Application class. You should inherit from this class to create your own custom
    pax25 apps.
    """

    def __init__(self, *, name: str, station: "Station", settings: S):
        self.name = name
        self.settings = settings
        self.connection_state_table: BasicStateTable = {}
        self.station = station
        self.setup()

    def setup(self) -> None:
        """
        Perform any initial state configuration for your application in this
        function.
        """

    def is_admin(self, connection: "Connection") -> bool:
        """
        Check if the current user is an admin.
        """
        return connection.is_admin and connection.first_party.name == self.station.name

    def on_connect(self, connection: "Connection") -> None:
        """
        Called when a new connection is established.
        """
        self.connection_state_table[connection] = {"command": b""}
        self.on_startup(connection)

    def on_startup(self, connection: "Connection") -> None:
        """
        Run right after a new connection is established. You can use this function to
        do any initial state configuration and/or send a welcome message.
        """

    def on_disconnect(self, connection: "Connection") -> None:
        """
        Run when a connection is being dropped.
        """
        self.on_shutdown(connection)
        del self.connection_state_table[connection]

    def on_shutdown(self, connection: "Connection") -> None:
        """
        Called when a connection is being disconnected. Perform any cleanup here.
        """

    def on_message(self, connection: "Connection", message: str) -> None:
        """
        Called when a message is received. By default, this is called by
        on_bytes when it detects a carriage return has been sent.
        """

    def on_bytes(
        self,
        connection: "Connection",
        bytes_received: bytes,
    ) -> None:
        """
        Called when bytes are received from a connection for this application. You
        usually don't want to call this directly, but you might need to if you need
        to control how bytes sent from the client are handled.
        """
        # Not sure how often we'll receive packets with carriage returns in the
        # middle of them, but for most applications that should indicate the
        # end of one command and the start of another, so we break them up here.
        for raw_int in bytes_received:
            # Iterating over bytes produces ints.
            byte = raw_int.to_bytes(1, sys.byteorder)
            try:
                self.connection_state_table[connection]["command"]
            except KeyError:
                # Disconnected between our previous processed byte and our current one.
                return
            match byte:
                case b"\r":
                    current_message = (
                        self.connection_state_table[connection]["command"] + byte
                    ).decode("utf-8")
                    # Clear the command before sending the message in case there's an
                    # exception.
                    self.connection_state_table[connection]["command"] = b""
                    self.on_message(
                        connection,
                        # Remove trailing newline.
                        current_message[:-1],
                    )
                # Backspace
                case b"\x7f":
                    current_string = self.connection_state_table[connection][
                        "command"
                    ].decode("utf-8", errors="backslashreplace")
                    self.connection_state_table[connection]["command"] = current_string[
                        :-1
                    ].encode("utf-8")
                case _:
                    self.connection_state_table[connection]["command"] += byte



"""
Data structures and functions used for ingesting and exporting KISS frames.
"""

import logging
from collections.abc import AsyncIterator, Awaitable, Callable
from dataclasses import dataclass
from typing import NamedTuple, Self, TypeVar, cast

from pax25.ax25.exceptions import DisassemblyError
from pax25.ax25.frame import Frame
from pax25.ax25.protocols import Assembler
from pax25.interfaces.kiss.constants import (
    KISS_CMD_DATA,
    KISS_CMD_PASSWORD,
    KISS_ENDIAN,
    KISS_ESCAPED_FEND,
    KISS_ESCAPED_FESC,
    KISS_FEND,
    KISS_FESC,
    KISS_SHIFT_PORT,
    KISS_TFEND,
    KISS_TFESC,
)

logger = logging.getLogger(__name__)


@dataclass
class ReaderState:
    """
    Dataclass for keeping track of KISS frame ingestion.
    """

    packet: bytes = b""
    command: bytes = b""
    packet_started: bool = False
    reading: bool = False
    escaped: bool = False
    ready: bool = False


ReaderUpdater = Callable[[bytes, ReaderState], None]


class APRSPasswordFrame(NamedTuple):
    call_sign: str
    password: int

    def size(self) -> int:
        """
        Return length of password frame.
        """
        return len(self.assemble())

    def assemble(self) -> bytes:
        """
        Generate a bytestring for sending over transport.
        """
        return (
            self.call_sign.encode("utf-8") + b"\r" + str(self.password).encode("utf-8")
        )

    @property
    def valid(self) -> bool:
        """
        Returns if this frame contains a valid APRS username/password combo.
        """
        try:
            return aprs_password(self.call_sign) == self.password
        except ValueError:
            return False

    @classmethod
    def disassemble(cls, data: bytes) -> Self:
        """
        Given a bytestring, construct an APRSPasswordFrame.
        """
        segments = data.split(b"\r", maxsplit=1)
        if len(segments) != 2:
            raise DisassemblyError("Password missing.")
        [raw_call_sign, raw_password] = segments
        try:
            # Note: We don't verify a call sign is real, just that it is a valid string.
            call_sign = raw_call_sign.decode("utf-8")
        except ValueError as err:
            raise DisassemblyError("Could not decode call sign.") from err
        try:
            password = int(raw_password.decode("utf-8"))
        except ValueError as err:
            raise DisassemblyError("Password is not an integer.") from err
        return cls(call_sign=call_sign, password=password)


class EmptyFrame(NamedTuple):
    """
    Used for frames that are intended to be "empty" since the command is sufficient.
    Contains one byte of data to avoid making parsing more complicated on the KISS-side.
    """

    def size(self) -> int:
        """
        Returns the length of the frame, which is 0 because it's empty.
        """
        return 0

    def assemble(self) -> bytes:
        """
        Returns a single byte, content doesn't matter.
        """
        return b""

    @classmethod
    def disassemble(cls, data: bytes) -> Self:
        """
        Returns an instance of the frame.
        """
        return cls()


class BadCredentialsFrame(EmptyFrame):
    """
    Used when the credentials received are bogus.
    """


def requires_read(func: ReaderUpdater) -> ReaderUpdater:
    """
    Makes a function inert if the 'reading' state isn't set
    to true.
    """

    def wrapped(byte: bytes, state: ReaderState) -> None:
        if not state.reading:
            return
        func(byte, state)

    return wrapped


def handle_f_end(_byte: bytes, state: ReaderState) -> None:
    """
    Handle an f-end marker.
    """
    if not state.command:
        state.packet_started = True
        return
    state.reading = False
    state.ready = True
    return


@requires_read
def handle_tf_esc(byte: bytes, state: ReaderState) -> None:
    """
    Handle a tf escape.
    """
    if state.escaped:
        state.packet += KISS_FESC
        state.escaped = False
        return
    # Just a normal byte if we're not escaped.
    state.packet += byte


@requires_read
def handle_f_esc(_byte: bytes, state: ReaderState) -> None:
    """
    Handle an f escape.
    """
    if state.escaped:
        # This is an error. Continue.
        state.escaped = False
        return
    state.escaped = True


@requires_read
def handle_tf_end(byte: bytes, state: ReaderState) -> None:
    """
    Handle a tf_end.
    """
    if state.escaped:
        state.packet += KISS_FEND
        state.escaped = False
        return
    # Not in escaped mode, so this is the actual byte.
    state.packet += byte


@requires_read
def handle_new_char(byte: bytes, state: ReaderState) -> None:
    """
    Handle any other byte character that's part of the KISS frame.
    """
    state.packet += byte


COMMAND_FUNCS: dict[bytes, ReaderUpdater] = {
    KISS_FEND: handle_f_end,
    KISS_FESC: handle_f_esc,
    KISS_TFEND: handle_tf_end,
    KISS_TFESC: handle_tf_esc,
}


def kiss_command(command: int, port: int, value: bytes) -> bytes:
    """
    Format a kiss command for transmission.
    """
    cmd_byte = port
    cmd_byte <<= KISS_SHIFT_PORT
    cmd_byte |= command
    value = value.replace(KISS_FESC, KISS_ESCAPED_FESC)
    value = value.replace(KISS_FEND, KISS_ESCAPED_FEND)
    return KISS_FEND + cmd_byte.to_bytes(1, KISS_ENDIAN) + value + KISS_FEND


type CommandToDisassembler = dict[bytes, Callable[[bytes], Assembler]]


# If we ever need new frame command types for some reason, we can register them here.
def default_frame_disassemblers() -> CommandToDisassembler:
    """
    Returns a default set of frame disassemblers for use with reading in KISS commands.
    """
    return {KISS_CMD_DATA.to_bytes(1, KISS_ENDIAN): Frame.disassemble}


FT = TypeVar("FT", bound=Assembler)


async def filter_kiss_frames(
    read_byte: Callable[[], Awaitable[bytes]],
    frame_types: set[type[FT]],
    get_disassemblers: Callable[
        [], CommandToDisassembler
    ] = default_frame_disassemblers,
) -> AsyncIterator[FT]:
    """
    Pulls specific frame types from KISS
    """
    async for _, frame in read_from_kiss(
        read_byte, get_disassemblers=get_disassemblers
    ):
        if frame.__class__ in frame_types:
            yield cast(FT, frame)


async def ax25_frames_from_kiss(
    read_byte: Callable[[], Awaitable[bytes]],
) -> AsyncIterator[Frame]:
    """
    Generator loop for reading KISS from a source. Drops all frames except for AX.25
    frames.
    """
    async for frame in filter_kiss_frames(read_byte, frame_types={Frame}):
        yield frame


async def pull_password_frame(
    read_byte: Callable[[], Awaitable[bytes]],
) -> APRSPasswordFrame | None:
    """
    Returns the first frame received if it is a password frame, otherwise return None.
    """
    password_disassemblers = {
        KISS_CMD_PASSWORD.to_bytes(1, KISS_ENDIAN): APRSPasswordFrame.disassemble,
    }
    async for _, frame in read_from_kiss(
        read_byte,
        get_disassemblers=lambda: {
            **default_frame_disassemblers(),
            **password_disassemblers,
        },
    ):
        match frame:
            case APRSPasswordFrame():
                return frame
            case _:
                return None
    return None


async def read_from_kiss(
    read_byte: Callable[[], Awaitable[bytes]],
    get_disassemblers: Callable[
        [], CommandToDisassembler
    ] = default_frame_disassemblers,
) -> AsyncIterator[tuple[int, Assembler]]:
    """
    Generator loop for reading KISS frames from a source.
    """
    disassemblers = get_disassemblers()
    state = ReaderState()
    while byte := await read_byte():
        if state.packet_started and not state.reading and byte != KISS_FEND:
            # This is the command byte.
            state.command = byte
            state.reading = True
            continue
        modifier: ReaderUpdater = COMMAND_FUNCS.get(byte, handle_new_char)
        modifier(byte, state)
        if state.ready:
            try:
                disassembler = disassemblers.get(state.command, None)
                if disassembler is None:
                    raise DisassemblyError(
                        "No existing frame disassembler for this command."
                    )
                frame = disassembler(state.packet)
                yield int.from_bytes(state.command, KISS_ENDIAN), frame
            except Exception as err:
                logger.debug("Packet decoding error: %s", (err,))
            state = ReaderState()


def aprs_password(call_sign: str) -> int:
    """
    Generates an APRS "password" from a call sign.
    """
    call_sign = call_sign.split("-", maxsplit=1)[0].upper()[:10]
    if not call_sign:
        raise ValueError("Call sign empty.")
    call_sign_bytes = call_sign.encode("utf-8")
    base_code = 0x73E2
    index = 0
    length = len(call_sign_bytes)
    # This is trickier than it appears. Yes, we have to do it this way.
    while index < length:
        base_code ^= int(call_sign_bytes[index]) << 8
        if (index + 1) < length:
            base_code ^= call_sign_bytes[index + 1]
        index += 2
    return base_code & 0x7FFF



"""
CommandRouter module. We use the command router to route commands for the command line
app. Other apps can use it, too, as it's not strictly bound to the command line app.
"""

from collections.abc import Callable
from dataclasses import dataclass
from typing import Any

from pax25.applications.autocomplete import AutocompleteDict
from pax25.applications.parsers import ParseError, string_parser
from pax25.applications.utils import send_message
from pax25.exceptions import ConfigurationError
from pax25.services.connection.connection import Connection

type CommandFunc[P] = Callable[[Connection, "CommandContext[P]"], None]

type CommandMap = dict[str, CommandFunc[object]]


@dataclass(frozen=True, kw_only=True)
class CommandSpec[P]:
    """
    Used for defining commands that are used by the command line app, or any other app
    that follows its conventions for commands.
    """

    command: str
    aliases: tuple[str, ...] = tuple()
    # A long description of the command and its usage.
    help: str
    # The function for this command, when run.
    function: CommandFunc[P]
    parser: Callable[["ParserSpec"], P]


@dataclass(frozen=True, kw_only=True)
class ParserSpec:
    # The full raw input line from the user.
    raw_input: str
    # The first segment of the raw input line.
    command: str
    # The remainder of the input line after the command.
    args: str
    connection: Connection
    # Sometimes, we may need to know what the settings of the CommandSpec were to parse.
    # For an example of this, check 'requires_full_command'.
    # However, we need to prevent full type reflection here since parser wrappers may
    # modify types along the way. See the 'optional' wrapper in the parsers module for
    # an example of this.
    command_spec: CommandSpec[Any]


@dataclass()
class CommandContext[P]:
    """
    Context handed to a command.
    """

    spec: CommandSpec[P]
    # The actual command as entered by the user, as interpreted by the command router.
    command: str
    args: P
    # The full raw input from the user, as a string.
    raw_input: str


def default_command_func(connection: Connection, context: CommandContext[str]) -> None:
    """
    Default command used by CommandRouter.
    """
    if not context.command:
        # User entered nothing. Do nothing in response. Works best when the
        # post_command_func is set to show a prompt.
        return
    send_message(connection, f"{repr(context.command)} is not a recognized command.")


default_command: CommandSpec[str] = CommandSpec(
    command="",
    help="",
    function=default_command_func,
    parser=string_parser,
)


def default_post_command_func(_connection: Connection) -> None:
    """
    Default 'Post command'. Does nothing. Overwrite this to provide your own command
    prompt, for instance.
    """


class CommandRouter:
    """
    Router object that allows us to quickly route to command functions based on a
    command string sent by a user.
    """

    def __init__(
        self,
        *,
        default: CommandSpec[Any] = default_command,
        post_command_func: Callable[[Connection], None] = default_post_command_func,
    ) -> None:
        """
        Creates an autocompleting command map that handles input and runs any matching
        command.
        """
        self.command_store: AutocompleteDict[CommandSpec[Any]] = AutocompleteDict()
        # Canonical listing of all command names, used for checking conflicts.
        self.command_set: set[str] = set()
        # Canonical listing of all aliases, used for checking conflicts.
        self.alias_set: set[str] = set()
        self.default = default
        # A 'post command' that runs after a command is complete.
        self.post_command = post_command_func

    @property
    def help_available(self) -> bool:
        """
        Returns a boolean indicating if there's a help command installed.
        """
        try:
            results = self.command_store["HELP"]
            # Help command is ambiguous if there's more than 1.
            # This could still be wrong if there is an entry named something like
            # 'helpmeplz' and it's not a help command. But we're going to discount that
            # possibility here. If it ever becomes a real problem we'll refactor this
            # to something more robust.
            return len(results) == 1
        except KeyError:
            return False

    def add(self, *args: CommandSpec[Any]) -> None:
        """
        Add commands to the command router.
        """
        for arg in args:
            command = arg.command.upper()
            aliases = set(alias.upper() for alias in arg.aliases)
            to_check = (command, *aliases)
            for entry in to_check:
                if entry in self.command_set or entry in self.alias_set:
                    existing_spec = self.command_store[entry]
                    raise ConfigurationError(
                        f"Found preexisting entry with conflicting name or "
                        f"aliases when adding spec {repr(arg)}. Conflicting "
                        f"entry was: {repr(existing_spec)}"
                    )
            for entry in to_check:
                self.command_store[entry] = arg
            self.command_set |= {command}
            self.alias_set |= aliases

    def remove(self, *args: CommandSpec[Any]) -> None:
        """
        Remove commands from the command router.
        """
        for arg in args:
            command = arg.command.upper()
            if command not in self.command_store.store:
                raise KeyError(f"Command does not exist, {repr(arg.command)}")
            if self.command_store.store[command] != arg:
                raise KeyError(
                    f"Command {repr(arg.command)} exists, but is for a different spec!"
                )
            del self.command_store[command]
            self.command_set -= {command}
            for alias in arg.aliases:
                del self.command_store[alias.upper()]
                self.alias_set -= {alias.upper()}

    def route(self, connection: Connection, raw_command: str) -> None:
        """
        Routes a user to a command function based on their selection, or gives them
        a hint otherwise.
        """
        segments = raw_command.split(maxsplit=1)
        try:
            first_segment = segments[0]
            command = first_segment.upper()
        except IndexError:
            # No command specified-- it was an empty string. Run the default.
            context = CommandContext(
                command="",
                args="",
                spec=self.default,
                raw_input="",
            )
            self.default.function(connection, context)
            self.post_command(connection)
            return
        args = ""
        if len(segments) == 2:
            args = segments[1]
        try:
            candidates = self.command_store[command]
        except KeyError:
            context = CommandContext(
                command=first_segment,
                args=args,
                spec=self.default,
                raw_input=raw_command,
            )
            self.default.function(connection, context)
            self.post_command(connection)
            return
        if len(candidates) > 1:
            possibilities = sorted(entry.command for entry in candidates)
            send_message(
                connection,
                "Ambiguous command. "
                f"Did you mean one of these?: {', '.join(possibilities)}",
            )
            return
        [spec] = candidates
        try:
            parser_spec = ParserSpec(
                raw_input=raw_command,
                command=command,
                args=args,
                connection=connection,
                command_spec=spec,
            )
            parsed_args = spec.parser(parser_spec)
            context = CommandContext(
                command=first_segment,
                args=parsed_args,
                spec=spec,
                raw_input=raw_command,
            )
        except ParseError as err:
            error_string = str(err)
            if self.help_available:
                error_string += f"\rTry: help {spec.command}"
            send_message(connection, error_string)
            self.post_command(connection)
            return
        spec.function(connection, context)
        self.post_command(connection)



"""
Trie for auto-complete style lookups. Useful for short commands and help lookups.
"""


class AutocompleteDict[V]:
    """
    An autocomplete dictionary. Lookups return a set of candidate values.

    This is not especially space efficient-- proper refactoring to a true Trie might be
    better if we expect to have a lot of entries.
    """

    def __init__(self) -> None:
        """
        Set up our internal state for Autocomplete lookups.
        """
        self.store: dict[str, V] = {}
        self._lookups: dict[str, set[str]] = {}

    def __setitem__(
        self,
        key: str,
        value: V,
    ) -> None:
        """
        Add an entry and populate its autocomplete lookups.
        """
        existing = False
        if key in self.store:
            existing = True
        self.store[key] = value
        if existing:
            # No need to create the lookups-- they're already here.
            return
        current = ""
        for char in key:
            current += char
            if current in self._lookups:
                self._lookups[current] |= {key}
            else:
                self._lookups[current] = {key}

    def __getitem__(self, key: str) -> set[V]:
        """
        Get a key from the AutoComplete dict.
        """
        # Pull exact matches immediately.
        if key in self.store:
            return {self.store[key]}
        if key not in self._lookups:
            raise KeyError(key)
        return {self.store[value] for value in self._lookups[key]}

    def __delitem__(self, key: str) -> None:
        """
        Remove a key from the AutoComplete dict.
        """
        if key not in self.store:
            raise KeyError(key)
        del self.store[key]
        current = ""
        for char in key:
            current += char
            self._lookups[current] -= {key}
            if not self._lookups[current]:
                del self._lookups[current]



"""
Defines types used by interfaces.
"""

from dataclasses import dataclass
from typing import (
    IO,
    TYPE_CHECKING,
    Any,
    Generic,
    Literal,
    NotRequired,
    Self,
    TypedDict,
    TypeVar,
)

from pax25.ax25.frame import Frame
from pax25.types import EmptyDict
from pax25.utils import smart_clone

if TYPE_CHECKING:  # pragma: no cover
    from pax25.services.connection.connection import Connection
    from pax25.station import Station


S = TypeVar("S", bound=dict[str, Any] | EmptyDict, default=Any)


class Interface(Generic[S]):
    """
    The interface base class, which defines the required functions for an interface.
    All the actual functions must be defined on the subclass-- this base class
    just raises NotImplementedError for all of them.

    Attributes:
        listening: bool flag to indicate whether the interface is up and
            listening for packets.
        name: str name of the interface as instantiated, for internal reference.
        station: Station the interface is initialized for.
        sudo: Whether this interface can be used for superuser connections.
        settings: Current settings on the interface.
    """

    name: str
    type: str
    station: "Station"
    _settings: S

    @property
    def listening(self) -> bool:
        """
        Returns a flag indicating that the interface is active and listening.
        """
        raise NotImplementedError  # pragma: no cover

    @property
    def gateway(self) -> bool:
        """
        Whether this interface can be used to connect to the outside world.
        """
        raise NotImplementedError  # pragma: no cover

    @property
    def sudo(self) -> bool:
        """
        Whether connections from this interface should automatically be given
        administrative access.
        """
        raise NotImplementedError  # pragma: no cover

    @property
    def settings(self) -> S:
        """
        Returns a copy of our settings that's (mostly) safe to manipulate.
        """
        return smart_clone(self._settings)

    @classmethod
    def install(cls, station: "Station", *, name: str, settings: S) -> Self:
        """
        Installs this interface on a station.
        """
        instance = cls(name=name, settings=settings, station=station)
        station.add_interface(name, instance)
        return instance

    async def reload_settings(self, settings: S) -> None:
        """
        Interfaces should be able to hot-reload settings.
        """
        raise NotImplementedError  # pragma: no cover

    def __init__(self, name: str, settings: S, station: "Station"):
        """
        Initialize the interface. The interface will be initialized with
        its name, settings, and the station it is being initialized for.

        Under what conditions to set sudo is up to you, but it is set to False by
        default. The sudo flag indicates whether this interface can be used for
        superuser connections.

        It does not automatically mean that connections on this interface will be
        treated as superuser connections, but the base Application class will consider
        a user a superuser if they are connected to an interface while its sudo flag is
        True, and their name matches the station's default name.
        """
        raise NotImplementedError  # pragma: no cover

    def send_frame(self, frame: Frame) -> None:
        """
        Send this frame out on this interface.

        **NOTE**: While this function is a synchronous function, the implementation must
        queue the frame for processing asynchronously. The reason is that if the frame
        is processed synchronously, it will be tied in with the same codepath that sent
        the frame. This can result in unpredictable behavior such as a connection
        sending out another frame before it has a chance to increase its frame counter.
        """
        raise NotImplementedError  # pragma: no cover

    def start(self) -> None:
        """
        Interfaces should implement a 'start' function. This function should create
        the read-loop in a non-blocking manner-- specifically, it should create an async
        loop. See example implementations for how this is done.
        """
        raise NotImplementedError  # pragma: no cover

    async def shutdown(self) -> None:
        """
        This handles any cleanup needed to bring this interface offline, closing
        whatever read loops are in effect.
        """
        raise NotImplementedError  # pragma: no cover


class FileSettings(TypedDict, total=False):
    """
    Settings for the file input.
    """

    input: str | IO[bytes]
    output: str | IO[bytes]
    source_name: str
    source_ssid: int
    destination_name: str
    destination_ssid: int
    auto_shutdown: bool
    pre_shutdown_sleep: int
    # If auto_shutdown is True, also closes the event loop after shutdown.
    stop_loop: bool
    sudo: bool


class ConnectionSpec(TypedDict):
    """
    Specification of an IP-connected target.
    """

    host: str
    port: NotRequired[int]


class TerminalNodeControllerSettings(TypedDict, total=False):
    """
    Settings for the TNC, set up when enabling KISS mode. Check the KISS documentation
    to learn about these settings:

    https://www.ax25.net/kiss.aspx
    """

    port: int
    tx_delay: int
    persist: int
    slot_time: int
    full_duplex: int


class SerialSettings(TypedDict, total=False):
    """
    Settings for serial input.
    """

    path: str
    baud_rate: int
    timeout: int
    write_timeout: int
    rtscts: bool
    dsrdtr: bool
    xonxoff: bool
    gateway: bool
    # Set this to make the TNC leave kiss mode and immediately shut down the station.
    # Works for Kantronics.
    exit_kiss: bool
    tnc_settings: TerminalNodeControllerSettings


class TCPSettings(TypedDict, total=False):
    """
    Settings for the TCP interface
    """

    # Inbound settings.
    allow_inbound: bool
    listening_address: str
    listening_port: int

    # Outbound settings.
    call_sign: str
    password: int
    connections: list[ConnectionSpec]

    gateway: bool
    sudo: bool


class DummySettings(TypedDict, total=False):
    """
    Settings for the dummy system.
    """

    gateway: bool
    sudo: bool


class FileInterfaceConfig(TypedDict):
    """
    Configuration for file interface.
    """

    type: Literal["file"]
    settings: FileSettings


class DummyInterfaceConfig(TypedDict):
    """
    Configuration for dummy interface.
    """

    type: Literal["dummy"]
    settings: DummySettings


class SerialInterfaceConfig(TypedDict):
    """
    Configuration for the serial interface.
    """

    type: Literal["serial"]
    settings: SerialSettings


class TCPInterfaceConfig(TypedDict):
    """
    Configuration for the TCP Interface.
    """

    type: Literal["tcp"]
    settings: TCPSettings


# Type used for tracking the tty settings we manipulate when we use stdin.
TerminalSettings = list[int | list[bytes | int]]


@dataclass(kw_only=True, frozen=True)
class ResolvedConnectionSpec:
    """
    Variant of ConnectionSpec where we've made sure both host and port are set.
    """

    host: str
    port: int


class LoopbackSettings(TypedDict):
    """
    Configuration for Loopback interface.
    """

    connection: "Connection"



"""
The Frame Router routes frames to relevant matching functions used to perform any
protocol work.
"""

from datetime import UTC, datetime
from logging import getLogger
from typing import TYPE_CHECKING

from pax25.ax25.frame import Frame
from pax25.ax25.matchers import MatchCall
from pax25.utils import LazyRepr

if TYPE_CHECKING:  # pragma: no cover
    from pax25.interfaces import Interface
    from pax25.station import Station


logger = getLogger(__name__)


class FrameRouter:
    """
    The Frame queue is a sort of router. It takes in frames from the varying interfaces
    and makes sure they go to their proper destination, whether that be an existing
    connection object or re-transcribed and sent out to another interface.
    """

    def __init__(self, *, station: "Station"):
        self.station = station
        self.matchers: dict[str, MatchCall] = {}
        self.last_transmission: None | datetime = None

    def process_frame(self, interface: "Interface", frame: Frame) -> None:
        """
        Interfaces call this function to put a new frame in the queue, to be interpreted
        as a received frame.

        This function must be resilient to avoid bringing down the station from one bug.
        """
        # Could change while iterating, since we might do something like register a
        # connection.
        if interface.type != "File":
            logger.debug(
                "Received from %s: %s",
                interface.name,
                LazyRepr(frame),
            )
        matchers = list(self.matchers.items())
        for key, match_call in matchers:
            try:
                if not match_call.matcher(frame, interface):
                    continue
            except Exception:
                logger.exception(
                    "Got error for matcher with key %s:",
                    repr(key),
                )
                continue
            try:
                match_call.notify(frame, interface)
            except Exception:
                logger.exception(
                    "Got error for notifier with key %s:",
                    repr(key),
                )
                continue

    def send_frame(
        self,
        interface: "Interface",
        frame: Frame,
        update_timestamp: bool = True,
    ) -> None:
        """
        Send a frame out on a specific interface. Also does some bookkeeping in the
        process, like checking if we're sending over a gateway and updating our
        .last_transmission stamp if so.

        In the future, it may be possible to filter outgoing packets, or otherwise
        listen for them.
        """
        if interface.name not in self.station.interfaces:
            # Interface was removed, and thus inert.
            logger.debug(
                "Dropping packet for %s (not in interface table): %s",
                (
                    interface.name,
                    LazyRepr(frame),
                ),
            )
            return
        if interface.gateway and update_timestamp:
            self.last_transmission = datetime.now(UTC)
        if interface.type != "File":
            logger.debug("Sending on %s: %s", interface.name, LazyRepr(frame))
        interface.send_frame(frame)

    def register_matcher(self, key: str, match_call: MatchCall) -> None:
        """
        Registers a matcher.
        """
        if key in self.matchers:
            logger.warning(
                "Existing matcher for key %s, %s, was replaced with new matcher %s. "
                "This may be a collision or a sign of a cleanup issue.",
                repr(key),
                self.matchers[key],
                match_call,
            )
        self.matchers[key] = match_call

    def remove_matcher(self, key: str) -> None:
        """
        Removes a matcher based on its key.
        """
        if key not in self.matchers:
            logger.warning(
                "Non-existent key removed, %s. This may indicate that the key was "
                "never registered, or cleanup has been called multiple times "
                "unnecessarily.",
                repr(key),
            )
            return
        del self.matchers[key]



"""
Contains classes for managing state variables used in AX.25 connections.
"""

from collections.abc import Iterator
from dataclasses import dataclass, field

from pax25.ax25.constants import AX25_SEQ_MAX
from pax25.ax25.frame import Frame


class RolloverCounter:
    """
    Rollover value counter. Allows us to count up or down while rolling over a specific
    range of values.
    """

    def __init__(
        self,
        *,
        name: str,
        initial: int = 0,
        minimum: int = 0,
        maximum: int = AX25_SEQ_MAX,
    ):
        """
        Initializes the rollover counter.
        """
        self.name = name
        self.value = initial
        self.minimum = minimum
        self.maximum = maximum
        assert self.minimum <= self.value <= maximum, (
            f"Value {self.value} is not between {self.minimum} and {self.maximum}."
        )

    def __str__(self) -> str:
        """
        Represents the RolloverCounter as a string.
        """
        return f"Rollover Counter ({self.name}), value: {self.value}"

    def retrace_from(self, start_value: int) -> Iterator[int]:
        """
        Returns an iterable that returns all integers from a starting position to the
        current value, inclusive.
        """
        assert self.minimum <= start_value <= self.maximum
        current_value = self.value
        counting_value = start_value
        yield counting_value
        if counting_value == current_value:
            return
        while counting_value != current_value:
            counting_value = self.after(counting_value)
            if counting_value == current_value:
                # We'll send the final included value after the loop.
                break
            yield counting_value
        yield current_value

    def iterate_up_to(self, end_value: int) -> Iterator[int]:
        """
        Iterably advance this counter up to the target number. There are some tricky
        parts to this.

        Since we're using this to keep track of all unacknowledged frames, we could have
        eight valid values, but only 7 alternative values from the current one. So,
        if we get the current value, we haven't actually changed, and should return no
        items in iteration. If we end up with a value one previous to the current value,
        we would loop all the way around, minus actually including the current value.

        As we loop, we also set the current value as the new value.
        """
        assert self.minimum <= end_value <= self.maximum
        while self.value != end_value:
            self.increment()
            yield self.value

    def next(self) -> int:
        """
        Gets the next value that would be set if incrementing.
        """
        return self.after(self.value)

    def previous(self) -> int:
        """
        Gets the previous value that would be set if decrementing.
        """
        return self.before(self.value)

    def after(self, value: int) -> int:
        """
        Get the value that would come after the given one.
        """
        assert self.minimum <= value <= self.maximum
        value += 1
        if value > self.maximum:
            return self.minimum
        return value

    def before(self, value: int) -> int:
        """
        Get the value that would come before the given one.
        """
        assert self.minimum <= value <= self.maximum
        value -= 1
        if value < self.minimum:
            return self.maximum
        return value

    def increment(self) -> None:
        """
        Increment the value.
        """
        self.value = self.next()

    def decrement(self) -> None:
        """
        Decrement the value.
        """
        self.value = self.previous()


@dataclass
class FrameStateTracker:
    """
    Tracks the frame variables and sequence numbers as defined in the AX.25
    Specification.
    """

    vs: RolloverCounter = field(
        default_factory=lambda: RolloverCounter(name="Send State Variable")
    )
    vr: RolloverCounter = field(
        default_factory=lambda: RolloverCounter(name="Receive State Variable")
    )
    # Acknowledgement has to start at the maximum so the initial zeroth frame will be
    # forward.
    va: RolloverCounter = field(
        default_factory=lambda: RolloverCounter(
            name="Acknowledge State Variable", initial=AX25_SEQ_MAX
        )
    )
    # Not a rollover counter like the others. This is keeping track of how many frames
    # we've received without sending an acknowledgement of some kind. If this hits
    # a high enough amount, we need to send an RR manually to avoid having the
    # connection hang too long.
    received_without_acknowledgement: int = 0
    # Whether we've made a request to retransmit a frame. There can only be one
    # outstanding retransmission request at a time.
    request_retransmit: bool = False
    # We set this true when the other station has indicated that it is busy and unable
    # to accept more frames.
    other_station_busy: bool = False
    outstanding_frames: dict[int, Frame | None] = field(default_factory=dict)
    # Maximum length of a frame in bytes. The primary constraint is the length of the
    # info field, which must shrink based on the length of the header.
    maximum_transmission_unit: int = 256



"""
Utility commands, as opposed to utility functions, which would be put in a file
named 'utils' instead.
"""

import itertools

from pax25 import Station
from pax25.applications.parsers import no_arguments
from pax25.applications.router import CommandContext, CommandSpec
from pax25.applications.utils import build_columns, send_message
from pax25.services.connection.connection import Connection
from pax25.utils import gateways_for


def build_whoami_command() -> CommandSpec[None]:
    def whoami(connection: Connection, _context: CommandContext[None]) -> None:
        """
        Tells the user who they are.
        """
        send_message(connection, str(connection.first_party))

    return CommandSpec(
        command="whoami",
        help="Tells you what your current connected station name is.",
        parser=no_arguments,
        function=whoami,
    )


def build_ports_command(station: Station) -> CommandSpec[None]:
    def ports(connection: Connection, _context: CommandContext[None]) -> None:
        """
        List all ports (interfaces) available for connecting outward.
        """
        headers = ["PORT", "ALIAS", "TYPE"]
        source_gateways = gateways_for(station)
        if not source_gateways:
            send_message(
                connection,
                "No gateways configured. You may not connect to outbound stations.",
            )
            return
        gateways = [
            (str(port.number), port.name, port.type)
            for port in source_gateways.values()
        ]
        columns = build_columns(itertools.chain(headers, *gateways), num_columns=3)
        send_message(connection, "\r".join(columns))

    return CommandSpec(
        command="ports",
        help="\r".join(
            [
                "Usage: PORTS",
                "",
                "Lists all available interfaces/ports for connecting to "
                "outbound stations, including aliases. Aliases are case-"
                "sensitive.",
            ]
        ),
        parser=no_arguments,
        function=ports,
    )



"""
TCP Interface for AX25 frames/connections
"""

import asyncio
import logging
from asyncio import (
    Queue,
    QueueShutDown,
    Server,
    StreamReader,
    StreamWriter,
    Task,
    wait_for,
)
from collections.abc import Callable
from dataclasses import dataclass
from typing import TYPE_CHECKING

from pax25.ax25.frame import Frame
from pax25.ax25.protocols import Assembler
from pax25.exceptions import ConfigurationError
from pax25.interfaces.kiss.constants import (
    KISS_CMD_DATA,
    KISS_CMD_PASSWORD,
    KISS_CMD_PASSWORD_REJECT,
    KISS_ENDIAN,
)
from pax25.interfaces.kiss.protocol import (
    APRSPasswordFrame,
    BadCredentialsFrame,
    CommandToDisassembler,
    default_frame_disassemblers,
    filter_kiss_frames,
    kiss_command,
    pull_password_frame,
)
from pax25.interfaces.types import Interface, ResolvedConnectionSpec, TCPSettings
from pax25.timer import Timer
from pax25.utils import cancel_all

if TYPE_CHECKING:  # pragma: no cover
    from pax25.station import Station

logger = logging.getLogger(__name__)

DEFAULT_PORT = 7773
DEFAULT_ADDRESS = "0.0.0.0"


@dataclass(kw_only=True, frozen=True)
class SocketHandles:
    """
    Struct for keeping track of data handles for a TCP stream.
    """

    reader: StreamReader
    writer: StreamWriter


class TCPInterface(Interface[TCPSettings]):
    """
    Interface for linking pax25 stations over TCP.
    """

    type = "TCP"

    def __init__(self, name: str, settings: TCPSettings, station: "Station") -> None:
        """
        Initialize the TCP Interface.
        """
        self.name = name
        self._settings = settings
        self.station = station
        self.connections: set[TCPConnection] = set()
        self.shutting_down = False
        self.server_task: Task[None] | None = None
        self.server: None | Server = None

    @property
    def listening(self) -> bool:
        """
        Returns true if we have any connections still operating.
        """
        return bool(
            any(
                connection
                for connection in self.connections
                if (connection.writer and not connection.writer.is_closing())
            )
            or self.server
        )

    @property
    def gateway(self) -> bool:
        """
        Returns True if this interface can be used to make outbound connections.
        """
        return self._settings.get("gateway", True)

    @property
    def sudo(self) -> bool:
        """
        Returns True if connections on this interface should be considered privileged.
        """
        return self._settings.get("sudo", False)

    async def reload_settings(self, settings: TCPSettings) -> None:
        """
        Reload TCP interface settings.
        """
        await self.shutdown()
        self._settings = settings
        self.shutting_down = False
        self.start()

    def connection_closed(self, connection: "TCPConnection") -> None:
        """
        Callback for closing a TCP connection.
        """
        if self.shutting_down:
            # Do nothing, this will be cleaned up elsewhere.
            return
        if connection.inbound or connection.force_close:
            self.connections.remove(connection)
            return
        # Need to retry.
        connection.initialize_connection(delay=5)

    def handle_client(self, reader: StreamReader, writer: StreamWriter) -> None:
        """
        Set up a connection for an inbound client.
        """
        result = writer.transport.get_extra_info("peername")
        if result is None:  # pragma: no cover
            # Something went wrong. Underlying C library should have provided this
            # metadata.
            writer.close()
        address, port = result
        self.connections.add(
            TCPConnection(
                remote_station=ResolvedConnectionSpec(
                    host=address,
                    port=port,
                ),
                reader=reader,
                writer=writer,
                close_callback=self.connection_closed,
                interface=self,
                inbound=True,
            ),
        )

    async def start_server(self) -> None:
        """
        Start listening for inbound connections.
        """
        if not self._settings.get("allow_inbound"):
            return
        try:
            self.server = await asyncio.start_server(
                self.handle_client,
                self._settings.get("listening_address", DEFAULT_ADDRESS),
                self._settings.get("listening_port", DEFAULT_PORT),
            )
            async with self.server:
                await self.server.serve_forever()
        finally:
            self.server = None

    def build_connections(self) -> None:
        """
        Create all the connections for the TCP Interface.
        """
        connections = self._settings.get("connections", [])
        if not connections:
            return
        call_sign = self._settings.get("call_sign", "")
        password = self._settings.get("password", 0)
        if not (call_sign and password):
            raise ConfigurationError(
                "TCP settings must include both call_sign and password.",
            )
        for value in self._settings.get("connections", []):
            self.connections.add(
                TCPConnection(
                    remote_station=ResolvedConnectionSpec(
                        host=value["host"], port=value["port"]
                    ),
                    close_callback=self.connection_closed,
                    credentials=APRSPasswordFrame(
                        call_sign=call_sign,
                        password=password,
                    ),
                    interface=self,
                    inbound=False,
                )
            )

    def send_frame(self, frame: Frame) -> None:
        """
        Send a frame out to all the connections which are listening.
        """
        for connection in self.connections:
            connection.send_frame(frame)

    def start(self) -> None:
        """
        Set up the loops.
        """
        self.shutting_down = False
        self.build_connections()
        self.server_task = asyncio.ensure_future(self.start_server())

    async def shutdown(self) -> None:
        """
        Close out all connections.
        """
        self.shutting_down = True
        if self.server:
            # Must explicitly close clients so read/write file handles raise exceptions.
            # Otherwise, serve_forever could get deadlocked when cancelling if we're
            # mid-read. See https://github.com/python/cpython/issues/123720
            self.server.close_clients()
            self.server.close()
        await cancel_all([self.server_task])
        for connection in self.connections:
            await connection.close()
        self.server = None


class TCPConnection:
    """
    Represents a TCP Connection to another station.
    """

    remote_station: ResolvedConnectionSpec
    send_queue: Queue[tuple[Assembler, int]]
    inbound: bool

    def __init__(
        self,
        *,
        remote_station: ResolvedConnectionSpec,
        inbound: bool,
        interface: TCPInterface,
        close_callback: Callable[["TCPConnection"], None],
        reader: StreamReader | None = None,
        writer: StreamWriter | None = None,
        credentials: APRSPasswordFrame | None = None,
    ):
        """
        Set up the TCP connection tracker.
        """
        self._connect_future: Task[None] | None = None
        self._read_loop: Task[None] | None = None
        self._write_loop: Task[None] | None = None
        self.interface = interface
        # Used to indicate that we should not retry this connection.
        self.force_close = False
        self.remote_station = remote_station
        self.close_callback = close_callback
        self.send_queue: Queue[tuple[Assembler, int]] = Queue()
        self.inbound = inbound
        self.reader = reader
        self.writer = writer
        self.credentials = credentials
        self.retry_timer = Timer(name="retry_connection")
        if inbound:
            if not (reader and writer):
                raise ValueError(
                    "Inbound connections should be initialized with read and "
                    "write streams."
                )
            self.start_loops()
        else:
            if not self.credentials:
                raise ValueError(
                    "Outbound connections should be initialized with APRS credentials."
                )
            self.initialize_connection()

    def send_frame(self, frame: Assembler, command: int = KISS_CMD_DATA) -> None:
        """
        Sends a frame to this TCP connection.
        """
        self.send_queue.put_nowait((frame, command))

    async def read_loop(self) -> None:
        """
        Read loop for serial interface.
        """
        reader = self.reader
        assert reader
        if self.inbound:
            try:
                frame = await wait_for(
                    pull_password_frame(lambda: reader.read(1)), timeout=3
                )
            except (ConnectionResetError, TimeoutError) as err:
                logger.info(
                    f"({self.interface.name}) Dropped remote "
                    f"({self.remote_station.host}, {self.remote_station.port}): "
                    f"{err.__class__.__name__}"
                )
                asyncio.ensure_future(self.close())
                return
            if not frame or not frame.valid:
                # Awaiting will kill Cancel this task before it finishes.
                self.send_frame(BadCredentialsFrame(), KISS_CMD_PASSWORD_REJECT)
                # This will close the connection as well, see the write loop.
                logger.info(
                    f"({self.interface.name}) Dropped remote "
                    f"({self.remote_station.host}, {self.remote_station.port}) "
                    f"connection: Incorrect APRS Authentication details."
                )
                self.send_queue.shutdown()
                return

        def get_disassemblers() -> CommandToDisassembler:
            return {
                **default_frame_disassemblers(),
                KISS_CMD_PASSWORD_REJECT.to_bytes(
                    1, KISS_ENDIAN
                ): BadCredentialsFrame.disassemble,
            }

        try:
            async for frame in filter_kiss_frames(
                lambda: reader.read(1),
                {Frame, BadCredentialsFrame},
                get_disassemblers=get_disassemblers,
            ):
                match frame:
                    case BadCredentialsFrame():
                        self.force_close = True
                        logger.error(
                            f"({self.interface.name}) Remote station "
                            f"{self.remote_station.host}:"
                            f"{self.remote_station.port} "
                            f"rejected credentials.",
                        )
                        asyncio.ensure_future(self.close())
                        return
                    case Frame():
                        self.interface.station.frame_router.process_frame(
                            self.interface, frame
                        )
        except ConnectionResetError:
            asyncio.ensure_future(self.close())

    async def write_loop(self) -> None:
        """
        Write loop for TCP connection.
        """
        try:
            while item := await self.send_queue.get():
                [frame, command] = item
                writer = self.writer
                assert writer
                try:
                    writer.write(kiss_command(command, 0, frame.assemble()))
                    await writer.drain()
                except OSError:
                    # Awaiting will kill Cancel this task before it finishes.
                    asyncio.ensure_future(self.close())
                    return
        except QueueShutDown:
            # Awaiting will kill Cancel this task before it finishes.
            asyncio.ensure_future(self.close())

    async def close(self) -> None:
        """
        Close the connection.
        """
        await cancel_all((self._connect_future, self._read_loop, self._write_loop))
        self.send_queue.shutdown(immediate=True)
        if self.writer:
            self.writer.close()
        self.writer, self.reader = None, None
        self.close_callback(self)

    @property
    def connected(self) -> bool:
        """
        Return if this connection is open.
        """
        return bool(self.reader and self.writer)

    @property
    def connecting(self) -> bool:
        """
        Return if this connection is still in progress. Returns False if connected
        or disconnected.
        """
        if self.inbound:
            # Not connecting, as this is inbound.
            return False
        return bool(self._connect_future and not self._connect_future.done())

    def initialize_connection(self, delay: int = 0) -> None:
        """
        Sets up a connection to the outbound system.
        """
        self._connect_future = asyncio.ensure_future(self._initialize_connection(delay))

    def start_loops(self) -> None:
        """
        Start up the loops for communicating back and forth.
        """
        self._read_loop = asyncio.ensure_future(self.read_loop())
        self._write_loop = asyncio.ensure_future(self.write_loop())

    async def _initialize_connection(self, delay: int = 0) -> None:
        """
        Internal process used by initialize_connection to perform its work.
        """
        assert self.credentials
        self.send_queue = Queue()
        ip = self.remote_station.host
        port = self.remote_station.port
        await asyncio.sleep(delay)
        future = asyncio.open_connection(ip, port)
        try:
            self.reader, self.writer = await asyncio.wait_for(future, timeout=5)
            logger.debug(f"({self.interface.name}) Connected to {ip}:{port}")
        except (TimeoutError, ConnectionRefusedError) as err:
            logger.error(
                f"({self.interface.name}) Failed connecting to {ip}:{port}. "
                f"{err.__class__.__name__}: {err}"
            )
            await self.close()
            return
        self.send_frame(self.credentials, KISS_CMD_PASSWORD)
        self.start_loops()



"""
Shim application. Used to shim between two connections, such as is required when
a connecting station wants to connect to a third party using us as the jump box.
"""

from pax25 import Application
from pax25.contrib.command.types import ShimSettings
from pax25.services.connection.connection import Connection


class ShimApplication(Application[ShimSettings]):
    """
    Shim application. Shuttles data over to the source application/connection.

    Unlike other applications which are instantiated station-wide, this one is
    instantiated on-need and per connection.
    """

    hidden = True

    def on_bytes(self, connection: "Connection", bytes_received: bytes) -> None:
        """
        Forward bytes to the proxy application.
        """
        self.settings.upstream_application.on_proxy_bytes(
            self.settings.proxy_connection,
            bytes_received,
        )

    def on_killed(self, connection: "Connection") -> None:
        """
        Inform the upstream application that this connection is closed.
        """
        self.settings.upstream_application.on_proxy_killed(
            self.settings.proxy_connection, connection
        )



"""
Commands for administrative tasks.
"""

import asyncio
import json
from pathlib import Path

from pax25 import Station
from pax25.applications.parsers import (
    call_sign_parser,
    file_path_parser,
    no_arguments,
    optional,
    requires_full_name,
    string_parser,
)
from pax25.applications.router import CommandContext, CommandSpec
from pax25.applications.utils import send_message
from pax25.ax25.address import Address
from pax25.services.connection.connection import Connection
from pax25.types import StationConfig
from pax25.utils import safe_save


def build_shutdown_command(station: Station) -> CommandSpec[None]:
    def shutdown(connection: Connection, _context: CommandContext[None]) -> None:
        """
        Shuts down the station.
        """
        send_message(connection, "Shutting down station.")
        asyncio.ensure_future(station.shutdown())

    return CommandSpec(
        command="shutdown",
        help="Shuts down the station",
        parser=requires_full_name(no_arguments),
        function=shutdown,
    )


def build_save_command(station: Station) -> CommandSpec[str | None]:
    def save(connection: Connection, context: CommandContext[str | None]) -> None:
        """
        Saves the station's configuration to a config file.
        """
        file_path = context.args or station.config_file_path
        if not file_path:
            send_message(
                connection, "Please provide a save file path, such as './config.json'"
            )
            return
        try:
            safe_save(path=Path(file_path), data=station.settings, debug=True)
        except OSError as err:
            send_message(connection, str(err))
        send_message(connection, "Saved.")

    return CommandSpec(
        command="save",
        help="Saves the current station configuration",
        parser=optional(string_parser),
        function=save,
    )


def build_name_command(station: Station) -> CommandSpec[Address | None]:
    async def update_settings(
        previous_name: str, connection: Connection, settings: StationConfig
    ) -> None:
        await station.reload_settings(settings)
        send_message(
            connection,
            f"Station name is now {settings['name']} (was {previous_name}.)",
        )

    def name(connection: Connection, context: CommandContext[Address | None]) -> None:
        """
        Sets the name of the station, or if no name is provided, returns it.
        """
        current_name = station.name
        if context.args is None:
            send_message(connection, current_name)
            return
        settings = station.settings
        settings["name"] = str(context.args)
        asyncio.ensure_future(update_settings(current_name, connection, settings))

    return CommandSpec(
        command="name",
        help="Retrieves (or sets) the name of the current station",
        parser=optional(call_sign_parser),
        function=name,
    )


def build_save_file_command(station: Station) -> CommandSpec[Path | None]:
    def save_file(connection: Connection, context: CommandContext[Path | None]) -> None:
        """
        Set the default configuration file for saving/loading the state of the station.
        """
        if context.args is None:
            if not station.config_file_path:
                send_message(connection, "None set")
                return
            send_message(connection, str(station.config_file_path))
            return
        old_path = station.config_file_path
        station.config_file_path = str(context.args)
        send_message(
            connection,
            f"Set to {repr(str(context.args))}. Was {repr(old_path)}",
        )

    return CommandSpec(
        command="setsave",
        help="Displays (or sets) the station configuration file path",
        parser=optional(file_path_parser),
        function=save_file,
    )


def build_load_command(station: Station) -> CommandSpec[Path | None]:
    async def reload_settings(connection: Connection, settings: StationConfig) -> None:
        await station.reload_settings(settings)
        send_message(
            connection,
            "Loaded.",
        )

    def load(connection: Connection, context: CommandContext[Path | None]) -> None:
        """
        Loads a configuration file, updating the station's settings.
        """
        if context.args is None:
            if not station.config_file_path:
                send_message(connection, "Config file not specified.")
                return
            path = Path(station.config_file_path)
        else:
            path = context.args
        try:
            with open(path) as file:
                settings = json.load(file)
        except (OSError, ValueError) as err:
            send_message(
                connection,
                f"Error when loading {repr(str(path))}. {err.__class__}: {err}",
            )
        asyncio.ensure_future(reload_settings(connection, settings))

    return CommandSpec(
        command="load",
        help="Load settings from the default file, or a specific file if specified.",
        parser=optional(file_path_parser),
        function=load,
    )



"""
Digipeater system. Will determine whether a digipeating frame is one we need to relay,
and if so, will queue up the digipeated frame with the correct flags marked.
"""

from typing import TYPE_CHECKING

from pax25.ax25.address import Address
from pax25.ax25.frame import Frame
from pax25.ax25.matchers import MatchCall, needs_repeat_from
from pax25.ax25.utils import repeated_for
from pax25.interfaces import Interface
from pax25.types import DigipeaterSettings
from pax25.utils import smart_clone

if TYPE_CHECKING:  # pragma: no cover
    from pax25.station import Station


class Digipeater:
    """
    Digipeater class. Manages digipeating functions. Only ever digipeats on the
    interface it hears from.

    This may need to be expanded later to allow for multiple digipeater conditions,
    but for now we assume a single digipeater per station, which responds on the main
    station name with SSID 0.
    """

    def __init__(
        self,
        station: "Station",
        settings: DigipeaterSettings | None,
    ) -> None:
        """Initializes the digipeater."""
        self.station = station
        self._settings = settings or DigipeaterSettings(enabled=True)
        matcher = needs_repeat_from(self.address)
        self.station.frame_router.register_matcher(
            "digipeater", MatchCall(matcher=matcher, notify=self.repeat)
        )

    @property
    def address(self) -> Address:
        """
        Gets the repeater address of the station.
        """
        return Address(name=self.station.name, ssid=0)

    async def reload_settings(self, settings: DigipeaterSettings) -> None:
        """
        Reloads the service with new settings.
        """
        # We don't have to restart anything, since this is used on-demand.
        self._settings = settings

    @property
    def enabled(self) -> bool:
        """
        Whether the Digipeater is enabled.
        """
        return self._settings.get("enabled", True)

    @property
    def settings(self) -> DigipeaterSettings:
        """
        Returns the current digipeater settings.
        """
        return smart_clone(self._settings)

    def repeat(self, frame: Frame, interface: Interface) -> None:
        """
        Performs digipeating for a matched frame.
        """
        if not self.enabled:
            return
        self.station.frame_router.send_frame(
            interface, repeated_for(self.address, frame)
        )

    def unregister(self) -> None:
        """
        Remove the digipeater from the matchers, effectively disabling it.
        """
        self.station.frame_router.remove_matcher("digipeater")
